"""DBus backed display management for Mutter"""

from collections import namedtuple

import dbus

from lutris.settings import DEFAULT_RESOLUTION_HEIGHT, DEFAULT_RESOLUTION_WIDTH
from lutris.util.log import logger

DisplayConfig = namedtuple("DisplayConfig", ("monitors", "name", "position", "transform", "primary", "scale"))


class Output:
    """Representation of a physical display output"""

    def __init__(self, output_info):
        self._output = output_info

    def __repr__(self):
        return "<Output: %s %s (%s)>" % (self.vendor, self.product, self.display_name)

    @property
    def output_id(self):
        """ID of the output"""
        return self._output[0]

    @property
    def winsys_id(self):
        """The low-level ID of this output (XID or KMS handle)"""
        return self._output[1]

    @property
    def current_crtc(self):
        """The CRTC that is currently driving this output,
        or -1 if the output is disabled
        """
        return self._output[2]

    @property
    def crtcs(self):
        """All CRTCs that can control this output"""
        return self._output[3]

    @property
    def name(self):
        """The name of the connector to which the output is attached (like VGA1 or HDMI)"""
        return self._output[4]

    @property
    def modes(self):
        """Valid modes for this output"""
        return [int(mode_id) for mode_id in self._output[5]]

    @property
    def clones(self):
        """Valid clones for this output, ie other outputs that can be assigned
        the same CRTC as this one; if you want to mirror two outputs that don't
        have each other in the clone list, you must configure two different
        CRTCs for the same geometry.
        """
        return self._output[6]

    @property
    def properties(self):
        """Other high-level properties that affect this output; they are not
        necessarily reflected in the hardware.
        """
        return self._output[7]

    @property
    def vendor(self):
        """Vendor name of the output"""
        return str(self._output[7]["vendor"])

    @property
    def product(self):
        """Product name of the output"""
        return str(self._output[7]["product"])

    @property
    def display_name(self):
        """A human readable name of this output, to be shown in the UI"""
        return str(self._output[7]["display-name"])

    @property
    def is_primary(self):
        """True if the output is the primary one"""
        return bool(self._output[7]["primary"])


class DisplayMode:
    """Representation of a screen mode (resolution, refresh rate)"""

    def __init__(self, mode_info):
        self.mode_info = mode_info

    def __str__(self):
        return "%sx%s@%s" % (self.width, self.height, self.frequency)

    def __repr__(self):
        return "<DisplayMode: %sx%s@%s>" % (self.width, self.height, self.frequency)

    @property
    def id(self):  # pylint: disable=invalid-name
        """ID of the mode"""
        return str(self.mode_info[0])

    @property
    def winsys_id(self):
        """the low-level ID of this mode"""
        return str(self.mode_info[1])

    @property
    def width(self):
        """width in physical pixels"""
        return int(self.mode_info[2])

    @property
    def height(self):
        """height in physical pixels"""
        return int(self.mode_info[3])

    @property
    def frequency(self):
        """refresh rate"""
        return str(self.mode_info[4])

    @property
    def flags(self):
        """mode flags as defined in xf86drmMode.h and randr.h"""
        return self.mode_info[5]


class CRTC:
    """A CRTC (CRT controller) is a logical monitor, ie a portion of the
    compositor coordinate space. It might correspond to multiple monitors, when
    in clone mode, but not that it is possible to implement clone mode also by
    setting different CRTCs to the same coordinates.
    """

    def __init__(self, crtc_info):
        self.crtc_info = crtc_info

    def __repr__(self):
        return "%s %s %s" % (self.id, self.geometry_str, self.current_mode)

    @property
    def id(self):  # pylint: disable=invalid-name
        """The ID in the API of this CRTC"""
        return str(self.crtc_info[0])

    @property
    def winsys_id(self):
        """the low-level ID of this CRTC
        (which might be a XID, a KMS handle or something entirely different)"""
        return self.crtc_info[1]

    @property
    def geometry_str(self):
        """Return a human readable representation of the geometry"""
        return "%dx%d%s%d%s%d" % (
            self.geometry[0],
            self.geometry[1],
            "" if self.geometry[2] < 0 else "+",
            self.geometry[2],
            "" if self.geometry[3] < 0 else "+",
            self.geometry[3],
        )

    @property
    def geometry(self):
        """The geometry of this CRTC
        (might be invalid if the CRTC is not in use)
        """
        return (int(self.crtc_info[2]), int(self.crtc_info[3]), int(self.crtc_info[4]), int(self.crtc_info[5]))

    @property
    def current_mode(self):
        """The current mode of the CRTC, or -1 if this CRTC is not used
        Note: the size of the mode will always correspond to the width
        and height of the CRTC"""
        return int(self.crtc_info[6])

    @property
    def current_transform(self):
        """The current transform (espressed according to the wayland protocol)"""
        return str(self.crtc_info[7])

    @property
    def transforms(self):
        """All possible transforms"""
        return str(self.crtc_info[8])

    @property
    def properties(self):
        """Other high-level properties that affect this CRTC;
        they are not necessarily reflected in the hardware.
        No property is specified in this version of the API.
        """
        return str(self.crtc_info[9])


class MonitorMode(DisplayMode):
    """Represents a mode given by a Monitor instance
    In addition to DisplayMode objects, this gives acces to the current scaling
    used and some additional properties like is_current.
    """

    @property
    def width(self):
        """width in physical pixels"""
        return int(self.mode_info[1])

    @property
    def height(self):
        """height in physical pixels"""
        return int(self.mode_info[2])

    @property
    def frequency(self):
        """refresh rate"""
        return str(self.mode_info[3])

    @property
    def scale(self):
        """scale preferred as per calculations"""
        return float(self.mode_info[4])

    @property
    def supported_scale(self):
        """scales supported by this mode"""
        return self.mode_info[5]

    @property
    def properties(self):
        """Additional properties"""
        return self.mode_info[6]

    @property
    def is_current(self):
        """Return True if the mode is the current one"""
        return "is-current" in self.properties


class Monitor:
    """A physical monitor"""

    def __init__(self, monitor):
        self._monitor = monitor

    def get_current_mode(self):
        """Return the current mode"""
        for mode in self.get_modes():
            if mode.is_current:
                return mode
        return

    def get_modes(self):
        """Return available modes"""
        return [MonitorMode(mode) for mode in self._monitor[1]]

    def get_mode_for_resolution(self, resolution):
        """Return an appropriate mode for a given resolution"""
        width, height = [int(i) for i in resolution.split("x")]
        for mode in self.get_modes():
            if mode.width == width and mode.height == height:
                return mode
        return

    @property
    def name(self):
        """Name of the connector"""
        return str(self._monitor[0][0])

    @property
    def vendor(self):
        """Manufacturer of the monitor"""
        return str(self._monitor[0][1])

    @property
    def model(self):
        """Model name of the monitor"""
        return str(self._monitor[0][2])

    @property
    def serial_number(self):
        """Serial number"""
        return str(self._monitor[0][3])

    @property
    def is_underscanning(self):
        """Return true if the monitor is underscanning"""
        return bool(self._monitor[2]["is-underscanning"])

    @property
    def is_builtin(self):
        """Return true if the display is builtin the machine (a laptop or a tablet)"""
        return bool(self._monitor[2]["is-builtin"])

    @property
    def display_name(self):
        """Human readable name of the display"""
        return str(self._monitor[2]["display-name"])


class LogicalMonitor:
    """A logical monitor. Similar to CRTCs but logical monitors also contain
    scaling information.
    """

    def __init__(self, lm_info, monitors):
        self._lm = lm_info
        self._monitors = monitors

    @property
    def position(self):
        """Return the position of the monitor"""
        return int(self._lm[0]), int(self._lm[1])

    @property
    def scale(self):
        """Scale"""
        return self._lm[2]

    @property
    def transform(self):
        """Transforms

        Possible transform values:
        0: normal
        1: 90°
        2: 180°
        3: 270°
        4: flipped
        5: 90° flipped
        6: 180° flipped
        7: 270° flipped
        """
        return self._lm[3]

    @property
    def primary(self):
        """True if this is the primary logical monitor"""
        return bool(self._lm[4])

    def _get_monitor_for_connector(self, connector):
        """Return a Monitor instance from its connector name"""
        for monitor in self._monitors:
            if monitor.name == str(connector):
                return monitor
        return

    @property
    def monitors(self):
        """Monitors displaying that logical monitor"""
        return [self._get_monitor_for_connector(m[0]) for m in self._lm[5]]

    @property
    def properties(self):
        """Possibly other properties"""
        return self._lm[6]

    def get_config(self):
        """Export the current configuration so it can be stored then reapplied later"""
        monitors = [(monitor.name, monitor.get_current_mode().id) for monitor in self.monitors]
        return DisplayConfig(monitors, self.monitors[0].name, self.position, self.transform, self.primary, self.scale)


class DisplayState:
    """Snapshot of a display configuration at a given time"""

    def __init__(self, interface):
        self.interface = interface
        self._state = self.load_state()

    def load_state(self):
        """Return current state from dbus interface"""
        return self.interface.GetCurrentState()

    @property
    def serial(self):
        """Configuration serial"""
        return self._state[0]

    @property
    def monitors(self):
        """Available monitors"""
        return [Monitor(monitor) for monitor in self._state[1]]

    @property
    def logical_monitors(self):
        """Current logical monitor configuration"""
        return [LogicalMonitor(l_m, self.monitors) for l_m in self._state[2]]

    @property
    def properties(self):
        """Display configuration properties"""
        return self._state[3]

    def get_primary_logical_monitor(self):
        """Return the primary logical monitor"""
        for lm in self.logical_monitors:
            if lm.primary:
                return lm
        return None

    def get_primary_monitor(self):
        """Returns the first physical monitor on the primary logical monitor."""
        lm = self.get_primary_logical_monitor()
        if lm:
            return lm.monitors[0]
        return self.monitors[0]

    def get_current_mode(self):
        """Return the current mode"""
        return self.get_primary_monitor().get_current_mode()


class MutterDisplayConfig:
    """Class to interact with the Mutter.DisplayConfig service"""

    namespace = "org.gnome.Mutter.DisplayConfig"
    dbus_path = "/org/gnome/Mutter/DisplayConfig"

    # Methods used in ApplyMonitorConfig
    VERIFY_METHOD = 0
    TEMPORARY_METHOD = 1
    PERSISTENT_METHOD = 2

    def __init__(self):
        session_bus = dbus.SessionBus()
        proxy_obj = session_bus.get_object(self.namespace, self.dbus_path)
        self.interface = dbus.Interface(proxy_obj, dbus_interface=self.namespace)
        self.resources = self.interface.GetResources()
        self.current_state = DisplayState(self.interface)

    @property
    def serial(self):
        """
        @serial is an unique identifier representing the current state
        of the screen. It must be passed back to ApplyConfiguration()
        and will be increased for every configuration change (so that
        mutter can detect that the new configuration is based on old
        state)
        """
        return self.resources[0]

    @property
    def crtcs(self):
        """
        A CRTC (CRT controller) is a logical monitor, ie a portion
        of the compositor coordinate space. It might correspond
        to multiple monitors, when in clone mode, but not that
        it is possible to implement clone mode also by setting different
        CRTCs to the same coordinates.

        The number of CRTCs represent the maximum number of monitors
        that can be set to expand and it is a HW constraint; if more
        monitors are connected, then necessarily some will clone. This
        is complementary to the concept of the encoder (not exposed in
        the API), which groups outputs that necessarily will show the
        same image (again a HW constraint).

        A CRTC is represented by a DBus structure with the following
        layout:
        * u ID: the ID in the API of this CRTC
        * x winsys_id: the low-level ID of this CRTC (which might
                    be a XID, a KMS handle or something entirely
                    different)
        * i x, y, width, height: the geometry of this CRTC
                                (might be invalid if the CRTC is not in
                                use)
        * i current_mode: the current mode of the CRTC, or -1 if this
                        CRTC is not used
                        Note: the size of the mode will always correspond
                        to the width and height of the CRTC
        * u current_transform: the current transform (espressed according
                            to the wayland protocol)
        * au transforms: all possible transforms
        * a{sv} properties: other high-level properties that affect this
                            CRTC; they are not necessarily reflected in
                            the hardware.
                            No property is specified in this version of the API.

        Note: all geometry information refers to the untransformed
        display.
        """
        return [CRTC(crtc) for crtc in self.resources[1]]

    @property
    def outputs(self):
        """
        An output represents a physical screen, connected somewhere to
        the computer. Floating connectors are not exposed in the API.
        An output is a DBus struct with the following fields:
        * u ID: the ID in the API
        * x winsys_id: the low-level ID of this output (XID or KMS handle)
        * i current_crtc: the CRTC that is currently driving this output,
                          or -1 if the output is disabled
        * au possible_crtcs: all CRTCs that can control this output
        * s name: the name of the connector to which the output is attached
                  (like VGA1 or HDMI)
        * au modes: valid modes for this output
        * au clones: valid clones for this output, ie other outputs that
                     can be assigned the same CRTC as this one; if you
                     want to mirror two outputs that don't have each other
                     in the clone list, you must configure two different
                     CRTCs for the same geometry
        * a{sv} properties: other high-level properties that affect this
                            output; they are not necessarily reflected in
                            the hardware.
                            Known properties:
                            - "vendor" (s): (readonly) the human readable name
                                            of the manufacturer
                            - "product" (s): (readonly) the human readable name
                                             of the display model
                            - "serial" (s): (readonly) the serial number of this
                                            particular hardware part
                            - "display-name" (s): (readonly) a human readable name
                                                  of this output, to be shown in the UI
                            - "backlight" (i): (readonly, use the specific interface)
                                               the backlight value as a percentage
                                               (-1 if not supported)
                            - "primary" (b): whether this output is primary
                                             or not
                            - "presentation" (b): whether this output is
                                                  for presentation only
                            Note: properties might be ignored if not consistenly
                            applied to all outputs in the same clone group. In
                            general, it's expected that presentation or primary
                            outputs will not be cloned.
        """
        return [Output(output) for output in self.resources[2]]

    @property
    def modes(self):
        """
        A mode represents a set of parameters that are applied to
        each output, such as resolution and refresh rate. It is a separate
        object so that it can be referenced by CRTCs and outputs.
        Multiple outputs in the same CRTCs must all have the same mode.
        A mode is exposed as:
        * u ID: the ID in the API
        * x winsys_id: the low-level ID of this mode
        * u width, height: the resolution
        * d frequency: refresh rate
        * u flags: mode flags as defined in xf86drmMode.h and randr.h

        Output and modes are read-only objects (except for output properties),
        they can change only in accordance to HW changes (such as hotplugging
        a monitor), while CRTCs can be changed with ApplyConfiguration().

        XXX: actually, if you insist enough, you can add new modes
        through xrandr command line or the KMS API, overriding what the
        kernel driver and the EDID say.
        Usually, it only matters with old cards with broken drivers, or
        old monitors with broken EDIDs, but it happens more often with
        projectors (if for example the kernel driver doesn't add the
        640x480 - 800x600 - 1024x768 default modes). Probably something
        that we need to handle in mutter anyway.
        """
        return [DisplayMode(mode) for mode in self.resources[3]]

    @property
    def max_screen_width(self):
        """Maximum width supported"""
        return self.resources[4]

    @property
    def max_screen_height(self):
        """Maximum height supported"""
        return self.resources[5]

    def get_mode_for_resolution(self, resolution):
        """Return an appropriate mode for a given resolution"""
        width, height = [int(i) for i in resolution.split("x")]
        for mode in self.modes:
            if mode.width == width and mode.height == height:
                return mode
        return

    def get_primary_logical_monitor(self):
        """Return the primary logical monitor"""
        return self.current_state.get_primary_logical_monitor()

    def get_primary_monitor(self):
        """Returns the first physical monitor on the primary logical monitor."""
        return self.current_state.get_primary_monitor()

    def apply_monitors_config(self, display_configs):
        """Set the selected display to the desired resolution"""
        # Reload resources
        if not display_configs:
            logger.error("No display config given, not applying anything")
            return
        self.resources = self.interface.GetResources()
        self.current_state = DisplayState(self.interface)
        monitors_config = [
            [
                config.position[0],
                config.position[1],
                dbus.Double(config.scale),
                dbus.UInt32(config.transform),
                config.primary,
                [
                    [dbus.String(str(display_name)), dbus.String(str(mode)), {}]
                    for display_name, mode in config.monitors
                ],
            ]
            for config in display_configs
        ]
        self.interface.ApplyMonitorsConfig(self.current_state.serial, self.TEMPORARY_METHOD, monitors_config, {})


class MutterDisplayManager:
    """Manage displays using the DBus Mutter interface"""

    def __init__(self):
        self.display_config = MutterDisplayConfig()

    def get_config(self):
        """Return the current configuration for each logical monitor"""
        return [logical_monitor.get_config() for logical_monitor in self.display_config.current_state.logical_monitors]

    def get_display_names(self):
        """Return display names of connected displays"""
        return [output.display_name for output in self.display_config.outputs]

    def get_resolutions(self):
        """Return available resolutions"""
        resolutions = ["%sx%s" % (mode.width, mode.height) for mode in self.display_config.modes]
        if not resolutions:
            logger.error("Could not generate resolution list")
            return ["%sx%s" % (DEFAULT_RESOLUTION_WIDTH, DEFAULT_RESOLUTION_HEIGHT)]
        return sorted(set(resolutions), key=lambda x: int(x.split("x")[0]), reverse=True)

    def get_current_resolution(self):
        """Return the current resolution for the primary display"""
        logger.debug("Retrieving current resolution")
        current_mode = self.display_config.current_state.get_current_mode()
        if not current_mode:
            logger.error("Could not retrieve the current display mode")
            return str(DEFAULT_RESOLUTION_WIDTH), str(DEFAULT_RESOLUTION_HEIGHT)
        return str(current_mode.width), str(current_mode.height)

    def set_resolution(self, resolution):
        """Change the current resolution"""
        if isinstance(resolution, str):
            monitor = self.display_config.get_primary_monitor()
            mode = monitor.get_mode_for_resolution(resolution)
            if not mode:
                logger.error("Could not find  valid mode for %s", resolution)
                return
            config = [DisplayConfig([(monitor.name, mode.id)], monitor.name, (0, 0), 0, True, 1.0)]
            self.display_config.apply_monitors_config(config)
        elif resolution:
            self.display_config.apply_monitors_config(resolution)
        else:
            return

        # Load a fresh config since the current one has changed
        self.display_config = MutterDisplayConfig()



import os

from lutris import settings
from lutris.util import http
from lutris.util.extract import extract_archive
from lutris.util.log import logger
from lutris.util.system import get_md5_hash

archive_formats = [".zip", ".7z", ".rar", ".gz"]
save_formats = [".srm"]
PLATFORM_PATTERNS = {
    "3DO": "3do",
    "Amiga CD32": "amiga-cd32",
    "Amiga": "amiga",
    "Master System": "sms",
    "Genesis": "md",
    "Game Gear": "gg",
    "Sega CD": "segacd",
    "Saturn": "saturn",
    "Dreamcast": "dc",
    "PICO": "pico",
    "ColecoVision": "colecovision",
    "Atari 8bit": "atari800",
    "Atari - 8-bit": "atari800",
    "Atari 2600": "atari2600",
    "Atari - 2600": "atari2600",
    "Atari Lynx": "lynx",
    "Atari ST": "atari-st",
    "Atari - ST": "atari-st",
    "Atari Jaguar": "jaguar",
    "Nintendo DS": "ds",
    "Super Nintendo Entertainment System": "snes",
    "Super Famicom": "snes",
    "Nintendo Famicom": "nes",
    "Nintendo Entertainment System": "nes",
    "Nintendo 64": "n64",
    "Game Boy Advance": "gba",
    "Game Boy": "gb",
    "GameCube": "gamecube",
    "Wii": "wii",
    "Switch": "switch",
    "PlayStation 2": "ps2",
    "PlayStation 3": "ps3",
    "PlayStation Vita": "psvita",
    "PlayStationPortable": "psp",
    "PlayStation Portable": "psp",
    "PlayStation": "ps1",
    "CD-i": "cdi",
    "MSX2": "msx",
    "Archimedes": "archimedes",
    "Acorn BBC": "bbc",
    "Acorn Electron": "electron",
    "Bally": "astrocade",
    "WonderSwan Color": "wonderswancolor",
    "WonderSwan": "wonderswan",
    "Amstrad CPC - Games - [DSK]": "cpc6128disk",
    "Amstrad CPC - Games - [CPR]": "gx4000",
    "Commodore C64 - Games - [PRG]": "c64prg",
    "Commodore C64 - Games - Arcade - [PRG]": "c64prg",
    "Commodore C64 - Games - [G64]": "c64g64",
    "Commodore C64 - Games - Arcade - [G64]": "c64g64",
    "ZX Spectrum - Games - [TZX]": "spectrumcass",
    "ZX Spectrum - Games - [TAP]": "spectrumcass",
    "Apple II": "apple2",
}


def search_tosec_by_md5(md5sum):
    """Retrieve a lutris bundle from the API"""
    if not md5sum:
        return []
    url = settings.SITE_URL + "/api/tosec/games?md5=" + md5sum
    response = http.Request(url, headers={"Content-Type": "application/json"})
    try:
        response.get()
    except http.HTTPError as ex:
        logger.error("Unable to get bundle from API: %s", ex)
        return None
    response_data = response.json
    return response_data["results"]


def scan_folder(folder, extract_archives=False):
    roms = {}
    archives = []
    saves = {}
    checksums = {}
    archive_contents = []
    if extract_archives:
        for filename in os.listdir(folder):
            basename, ext = os.path.splitext(filename)
            if ext not in archive_formats:
                continue
            extract_archive(os.path.join(folder, filename), os.path.join(folder, basename), merge_single=False)
            for archive_file in os.listdir(os.path.join(folder, basename)):
                archive_contents.append("%s/%s" % (basename, archive_file))

    for filename in os.listdir(folder) + archive_contents:
        basename, ext = os.path.splitext(filename)
        if ext in archive_formats:
            archives.append(filename)
            continue
        if ext in save_formats:
            saves[basename] = filename
            continue
        if os.path.isdir(os.path.join(folder, filename)):
            continue

        md5sum = get_md5_hash(os.path.join(folder, filename))
        roms[filename] = search_tosec_by_md5(md5sum)
        checksums[md5sum] = filename

    for rom, result in roms.items():
        if not result:
            print("no result for %s" % rom)
            continue
        if len(result) > 1:
            print("More than 1 match for %s", rom)
            continue
        print("Found: %s" % result[0]["name"])
        roms_matched = 0
        renames = {}
        for game_rom in result[0]["roms"]:
            source_file = checksums[game_rom["md5"]]
            dest_file = game_rom["name"]
            renames[source_file] = dest_file
            roms_matched += 1
        if roms_matched == len(result[0]["roms"]):
            for source, dest in renames.items():
                base_name, _ext = os.path.splitext(source)
                dest_base_name, _ext = os.path.splitext(dest)
                if base_name in saves:
                    save_file = saves[base_name]
                    _base_name, ext = os.path.splitext(save_file)
                    os.rename(os.path.join(folder, save_file), os.path.join(folder, dest_base_name + ext))
                try:
                    os.rename(os.path.join(folder, source), os.path.join(folder, dest))
                except FileNotFoundError:
                    logger.error("Failed to rename %s to %s", source, dest)


def guess_platform(game):
    category = game["category"]["name"]
    for pattern, platform in PLATFORM_PATTERNS.items():
        if pattern in category:
            return platform


def clean_rom_name(name):
    in_parens = False
    good_index = 0
    for i, c in enumerate(name[::-1], start=1):
        if c in (")", "]"):
            in_parens = True
        if in_parens:
            good_index = i
        if c in ("(", "]"):
            in_parens = False
    name = name[: len(name) - good_index].strip()
    if name.endswith(", The"):
        name = "The " + name[:-5]
    return name



import json
import time
from typing import List, Optional

from lutris import settings
from lutris.api import read_api_key
from lutris.database.categories import (
    add_category,
    add_game_to_category,
    get_all_games_categories,
    get_categories,
    remove_category_from_game,
)
from lutris.database.games import add_game, get_games, get_games_where
from lutris.game import Game
from lutris.gui.widgets import NotificationSource
from lutris.util import http
from lutris.util.log import logger

LIBRARY_URL = settings.SITE_URL + "/api/users/library"
LOCAL_LIBRARY_SYNCING = NotificationSource()
LOCAL_LIBRARY_SYNCED = NotificationSource()
LOCAL_LIBRARY_UPDATED = NotificationSource()
_IS_LOCAL_LIBRARY_SYNCING = False


def is_local_library_syncing():
    """True if the library is syncing now; attempting to sync again will do nothing if so."""
    # This provides access to the mutable global _IS_LOCAL_LIBRARY_SYNCING in a safer
    # way; if you just import the global directly you get a copy of its current state at import
    # time which is not very useful.
    return _IS_LOCAL_LIBRARY_SYNCING


class LibrarySyncer:
    def __init__(self):
        self.categories = self._load_categories()
        self.category_ids = self._load_categories(reverse=True)
        self.games_categories = get_all_games_categories()

    def _load_categories(self, reverse=False):
        """Create a mapping of category ID to name or name to ID if reverse is used"""
        key = "name" if reverse else "id"
        value = "id" if reverse else "name"
        return {r[key]: r[value] for r in get_categories()}

    def _get_request(self, since=None):
        """Return a request object and ensures authentication to the Lutris API"""
        credentials = read_api_key()
        if not credentials:
            return
        url = LIBRARY_URL
        if since:
            url += "?since=%s" % since
        return http.Request(
            url,
            headers={
                "Content-Type": "application/json",
                "Authorization": "Token " + credentials["token"],
            },
        )

    def _make_game_key(self, game):
        """Create a unique-ish key used to discriminate between games"""
        return (
            game["slug"],
            game["runner"] or "",
            game["platform"] or "",
            game["service"] or "",
        )

    def _get_game(self, remote_game) -> Optional[Game]:
        """Return a Game instance from a remote API record"""
        conditions = {"slug": remote_game["slug"]}
        for cond_key in ("runner", "platform", "service"):
            if remote_game[cond_key]:
                conditions[cond_key] = remote_game[cond_key]
        pga_game = get_games_where(**conditions)
        if len(pga_game) == 0:
            logger.error("No game found for %s", remote_game["slug"])
            return None
        if len(pga_game) > 1:
            logger.error("More than one game found for %s", remote_game["slug"])
            return None
        pga_game = pga_game[0]
        return Game(pga_game["id"])

    def _create_new_game(self, remote_game):
        """Create a new local game from a remote record"""
        logger.info("Create %s", remote_game["slug"])
        game_id = add_game(
            name=remote_game["name"],
            slug=remote_game["slug"],
            runner=remote_game["runner"],
            platform=remote_game["platform"],
            lastplayed=remote_game["lastplayed"],
            playtime=remote_game["playtime"],
            service=remote_game["service"],
            service_id=remote_game["service_id"],
            installed=0,
        )
        for category in remote_game["categories"]:
            self._ensure_category(category)
            add_game_to_category(game_id, self.category_ids[category])

    def _ensure_category(self, category):
        """Make sure a given category exists in the database, create it if not"""
        if category not in self.categories.values():
            add_category(category)
            self.categories = self._load_categories()
            self.category_ids = self._load_categories(reverse=True)

    def _update_categories(self, game: Game, remote_game: dict):
        """Update the categories of a local game"""
        game_categories: List[str] = game.get_categories()
        remote_categories: List[str] = remote_game["categories"]
        for category in game_categories:
            if category not in remote_categories:
                remove_category_from_game(game.id, self.category_ids[category])
        for category in remote_categories:
            if category not in game_categories:
                self._ensure_category(category)
                add_game_to_category(game.id, self.category_ids[category])

    def _db_game_to_api(self, db_game):
        """Serialize DB game entry to a payload compatible with the API"""
        categories = [self.categories[cat_id] for cat_id in self.games_categories.get(db_game["id"], [])]
        return {
            "name": db_game["name"],
            "slug": db_game["slug"],
            "runner": db_game["runner"] or "",
            "platform": db_game["platform"] or "",
            "playtime": "%0.5f" % (db_game["playtime"] or 0),
            "lastplayed": db_game["lastplayed"] or 0,
            "service": db_game["service"] or "",
            "service_id": db_game["service_id"] or "",
            "categories": categories,
        }

    def _db_games_to_api(self, db_games, since=None):
        """Serialize a collection of games to API format, optionally filtering by date"""
        payload = []
        for db_game in db_games:
            lastplayed = db_game["lastplayed"] or 0
            installed_at = db_game["installed_at"] or 0

            if since and lastplayed < since and installed_at < since:
                continue
            payload.append(self._db_game_to_api(db_game))
        return payload

    def sync_local_library(self, force: bool = False) -> None:
        """Sync task to send recent changes to the server and sync back server changes to the local client"""
        global _IS_LOCAL_LIBRARY_SYNCING

        if _IS_LOCAL_LIBRARY_SYNCING:
            return

        if not force and settings.read_setting("last_library_sync_at"):
            since = int(settings.read_setting("last_library_sync_at"))
        else:
            since = None
        all_games = get_games()
        local_library = self._db_games_to_api(all_games)
        local_library_updates = self._db_games_to_api(all_games, since=since)

        request = self._get_request(since)
        if not request:
            return

        LOCAL_LIBRARY_SYNCING.fire()
        any_local_changes = False
        try:
            _IS_LOCAL_LIBRARY_SYNCING = True
            try:
                request.post(data=json.dumps(local_library_updates).encode())
            except http.HTTPError as ex:
                logger.error("Could not send local library to server: %s", ex)
                return None
            library_keys = set()
            duplicate_keys = set()
            library_map = {}
            library_slugs = set()
            for game in local_library:
                library_key = self._make_game_key(game)
                if library_key in library_keys:
                    duplicate_keys.add(library_key)
                library_keys.add(library_key)
                library_map[library_key] = game
                library_slugs.add(game["slug"])

            for remote_game in request.json:
                remote_key = self._make_game_key(remote_game)
                if remote_key in duplicate_keys:
                    logger.warning("Duplicate game %s, not syncing.", remote_key)
                    continue
                if remote_key in library_map:
                    game = self._get_game(remote_game)
                    if not game:
                        continue
                    changed = False
                    if remote_game["playtime"] > game.playtime:
                        game.playtime = remote_game["playtime"]
                        changed = True
                    if remote_game["lastplayed"] > game.lastplayed:
                        game.lastplayed = remote_game["lastplayed"]
                        changed = True
                    if set(remote_game["categories"]) != set(game.get_categories()):
                        self._update_categories(game, remote_game)
                    if changed:
                        any_local_changes = True
                        game.save()
                else:
                    if remote_game["slug"] in library_slugs:
                        continue
                    self._create_new_game(remote_game)
                    any_local_changes = True

            settings.write_setting("last_library_sync_at", int(time.time()))
        finally:
            _IS_LOCAL_LIBRARY_SYNCING = False
            LOCAL_LIBRARY_SYNCED.fire()
            if any_local_changes:
                LOCAL_LIBRARY_UPDATED.fire()

    def delete_from_remote_library(self, games):
        """Task to delete a game entry from the remote library"""
        request = self._get_request()
        if not request:
            return
        try:
            request.delete(data=json.dumps(self._db_games_to_api(games)).encode())
        except http.HTTPError as ex:
            logger.error(ex)
            return None
        return request.json



"""Lutris installer class"""

import json
from functools import cached_property
from gettext import gettext as _
from pathlib import Path

from lutris.config import LutrisConfig, write_game_config
from lutris.database.games import add_or_update, get_game_by_field
from lutris.exceptions import AuthenticationError, UnavailableGameError
from lutris.installer import AUTO_ELF_EXE, AUTO_WIN32_EXE
from lutris.installer.errors import ScriptingError
from lutris.installer.installer_file import InstallerFile
from lutris.runners import import_runner
from lutris.services import SERVICES
from lutris.util.game_finder import find_linux_game_executable, find_windows_game_executable
from lutris.util.gog import convert_gog_config_to_lutris, get_gog_config_from_path, get_gog_game_path
from lutris.util.log import logger
from lutris.util.moddb import ModDB, is_moddb_url
from lutris.util.system import fix_path_case


class LutrisInstaller:  # pylint: disable=too-many-instance-attributes
    """Represents a Lutris installer"""

    def __init__(self, installer, interpreter, service, appid):
        self.interpreter = interpreter
        self.installer = installer
        self.is_update = False

        try:
            self.version = installer["version"]
            self.slug = installer["slug"]
            self.year = installer.get("year")
            self.runner = installer["runner"]
            self.script = installer.get("script")
            self.game_name = installer["name"]
            self.game_slug = installer["game_slug"]
        except KeyError as ex:
            raise ScriptingError(_("The script was missing the '%s' key, which is required.") % ex.args[0]) from ex

        self.service = self.get_service(initial=service)
        self.service_appid = self.get_appid(installer, initial=appid)
        self.variables = self.script.get("variables", {})
        self.script_files = [
            InstallerFile(self.game_slug, file_id, file_meta)
            for file_desc in self.script.get("files", [])
            for file_id, file_meta in file_desc.items()
        ]
        self.files = []
        self.extra_file_paths = []
        self.requires = self.script.get("requires")
        self.extends = self.script.get("extends")
        self.game_id = self.get_game_id()
        self.is_gog = False
        self.discord_id = installer.get("discord_id")

    @cached_property
    def scriptdir(self):
        installer_file = self.installer.get("installer_file")
        if not installer_file:
            return ""
        file = Path(installer_file).resolve()
        if not file.exists():
            return ""
        return str(file.parent) if file.is_file() else str(file)

    def get_service(self, initial=None):
        if initial:
            return initial
        if "steam" in self.runner and "steam" in SERVICES:
            return SERVICES["steam"]()
        version = self.version.lower()
        if "humble" in version and "humblebundle" in SERVICES:
            return SERVICES["humblebundle"]()
        if "gog" in version and "gog" in SERVICES:
            return SERVICES["gog"]()
        if "itch.io" in version and "itchio" in SERVICES:
            return SERVICES["itchio"]()

    def get_appid(self, installer, initial=None):
        if installer.get("is_dlc"):
            return installer.get("dlcid")
        if initial:
            return initial
        if not self.service:
            return
        service_id = None
        if self.service.id == "steam":
            service_id = installer.get("steamid") or installer.get("service_id")
        game_config = self.script.get("game", {})
        if self.service.id == "gog":
            service_id = game_config.get("gogid") or installer.get("gogid") or installer.get("service_id")
        if self.service.id == "humblebundle":
            service_id = game_config.get("humbleid") or installer.get("humblestoreid") or installer.get("service_id")
        if self.service.id == "itchio":
            service_id = game_config.get("itchid") or installer.get("itchid") or installer.get("service_id")
        if service_id:
            return service_id
        return

    @property
    def script_pretty(self):
        """Return a pretty print of the script"""
        return json.dumps(self.script, indent=4)

    def get_game_id(self):
        """Return the ID of the game in the local DB if one exists"""
        # If the game is in the library and uninstalled, the first installation
        # updates it
        existing_game = get_game_by_field(self.game_slug, "slug")
        if existing_game and (self.extends or not existing_game["installed"]):
            return existing_game["id"]

    @property
    def creates_game_folder(self):
        """Determines if an install script should create a game folder for the game"""
        if self.requires or self.extends:
            # Game is an extension of an existing game, folder exists
            return False
        if self.runner == "steam":
            # Steam games installs in their steamapps directory
            return False
        if not self.script.get("installer"):
            # No command can affect files
            return False
        if self.script_files or self.script.get("game", {}).get("gog") or self.script.get("game", {}).get("prefix"):
            return True
        command_names = [self.interpreter._get_command_name_and_params(c)[0] for c in self.script.get("installer", [])]
        if "insert_disc" in command_names:
            return True
        return False

    def get_errors(self):
        """Return potential errors in the script"""
        errors = []
        if not isinstance(self.script, dict):
            errors.append("Script must be a dictionary")
            # Return early since the method assumes a dict
            return errors

        # Check that installers contains all required fields
        for field in ("runner", "game_name", "game_slug"):
            if not hasattr(self, field) or not getattr(self, field):
                errors.append("Missing field '%s'" % field)

        # Check that libretro installers have a core specified
        if self.runner == "libretro":
            if "game" not in self.script or "core" not in self.script["game"]:
                errors.append("Missing libretro core in game section")

        # Check that Steam games have an AppID
        if self.runner == "steam":
            if not self.script.get("game", {}).get("appid"):
                errors.append("Missing appid for Steam game")

        # Check that installers don't contain both 'requires' and 'extends'
        if self.script.get("requires") and self.script.get("extends"):
            errors.append("Scripts can't have both extends and requires")
        return errors

    def prepare_game_files(self, extras, patch_version=None):
        """Gathers necessary files before iterating through them."""
        if not self.script_files:
            return

        installer_file_id = None
        installer_file_url = None
        if self.service:
            for file in self.script_files:
                if file.url.startswith("N/A"):
                    installer_file_id = file.id
                    installer_file_url = file.url
                    break
        files = [file.copy() for file in self.script_files if file.id != installer_file_id]
        extra_file_paths = []

        # Run variable substitution on the URLs from the script
        for file in files:
            file.set_url(self.interpreter._substitute(file.url))
            if is_moddb_url(file.url):
                file.set_url(ModDB().transform_url(file.url))

        if installer_file_id and self.service:
            logger.info("Getting files for %s", installer_file_id)
            try:
                if patch_version:
                    # If a patch version is given download the patch files instead of the installer
                    installer_files = self.service.get_patch_files(self, installer_file_id)
                else:
                    content_files, extra_files = self.service.get_installer_files(self, installer_file_id, extras)
                    extra_file_paths = [path for f in extra_files for path in f.get_dest_files_by_id().values()]
                    installer_files = content_files + extra_files
            except (AuthenticationError, UnavailableGameError) as ex:
                logger.exception("Game not available: %s", ex)
                installer_files = None

            if installer_files:
                for installer_file in installer_files:
                    files.append(installer_file)
            else:
                # Failed to get the service game, put back a user provided file
                logger.debug("Unable to get files from service. Setting %s to manual.", installer_file_id)
                files.insert(
                    0, InstallerFile(self.game_slug, installer_file_id, {"url": installer_file_url, "filename": ""})
                )

        # Commit changes only at the end; this is more robust in this method is runner
        # my two threads concurrently- the GIL can probably save us. It's not desirable
        # to do this, but this is the easiest workaround.
        self.files = files
        self.extra_file_paths = extra_file_paths

    def install_extras(self):
        # Copy extras to game folder; this updates the installer script, so it needs
        # be called just once, before launching the installers commands.
        if self.extra_file_paths and len(self.extra_file_paths) == len(self.files):
            # Reset the install script in case there are only extras.
            logger.warning("Installer with only extras and no game files")
            self.script["installer"] = []

        for extra_file in self.extra_file_paths:
            self.script["installer"].append({"copy": {"src": extra_file, "dst": "$GAMEDIR/extras"}})

    def _substitute_config(self, script_config):
        """Substitute values such as $GAMEDIR in a config dict."""
        config = {}
        for key in script_config:
            if not isinstance(key, str):
                raise ScriptingError(_("Game config key must be a string"), key)
            value = script_config[key]
            if str(value).lower() == "true":
                value = True
            if str(value).lower() == "false":
                value = False
            if key == "launch_configs":
                config[key] = [{k: self.interpreter._substitute(v) for (k, v) in _conf.items()} for _conf in value]
            elif isinstance(value, list):
                config[key] = [self.interpreter._substitute(i) for i in value]
            elif isinstance(value, dict):
                config[key] = {k: self.interpreter._substitute(v) for (k, v) in value.items()}
            elif isinstance(value, bool):
                config[key] = value
            else:
                config[key] = self.interpreter._substitute(value)
        return config

    def get_game_config(self):
        """Return the game configuration"""
        if self.requires:
            # Load the base game config
            required_game = get_game_by_field(self.requires, field="installer_slug")
            if not required_game:
                required_game = get_game_by_field(self.requires, field="slug")
            if not required_game:
                raise ValueError("No game matched '%s' on installer_slug or slug" % self.requires)
            base_config = LutrisConfig(runner_slug=self.runner, game_config_id=required_game["configpath"])
            config = base_config.game_level
        else:
            config = {"game": {}}

        # Config update
        if "system" in self.script:
            config["system"] = self._substitute_config(self.script["system"])
        if self.script.get(self.runner):
            installer_runner_config = self._substitute_config(self.script[self.runner])
            import_runner(self.runner)().adjust_installer_runner_config(installer_runner_config)
            config[self.runner] = installer_runner_config

        game_config = config["game"]

        entry_point_keys = ("iso", "rom", "main_file", "exe")

        if "game" in self.script:
            try:
                game_config.update(self.script["game"])
            except ValueError as err:
                raise ScriptingError(_("Invalid 'game' section"), faulty_data=self.script["game"]) from err

        # Obsolete install scripts may have the entry point key at root level;
        # we'll move them into the game-config if so, and if they are not already
        # there. Add a warning because I'm sure this compatibility ship will get lost,
        # and the scripts would be better updated.
        for entry_point_key in entry_point_keys:
            if entry_point_key in self.script and entry_point_key not in game_config:
                logger.warning("Moving entry point '%s' from script root level to the game config", entry_point_key)
                game_config[entry_point_key] = self.script[entry_point_key]

        game_config = self._substitute_config(game_config)
        if AUTO_ELF_EXE in game_config.get("exe", ""):
            game_config["exe"] = find_linux_game_executable(self.interpreter.target_path, make_executable=True)
        elif AUTO_WIN32_EXE in game_config.get("exe", ""):
            game_config["exe"] = find_windows_game_executable(self.interpreter.target_path)

        # Fix possible case differences
        for key in entry_point_keys:
            entry_point = game_config.get(key)
            if entry_point:
                game_config[key] = fix_path_case(entry_point)

        config["game"] = game_config
        config["name"] = self.game_name
        config["script"] = self.script
        config["variables"] = self.variables
        config["version"] = self.version
        config["requires"] = self.requires
        config["slug"] = self.slug
        config["game_slug"] = self.game_slug
        config["year"] = self.year
        if self.service:
            config["service"] = self.service.id
            config["service_id"] = self.service_appid
        return config

    def save(self):
        """Write the game configuration in the DB and config file"""
        if self.extends:
            logger.info(
                "This is an extension to %s, not creating a new game entry",
                self.extends,
            )
            return self.game_id

        if self.is_gog:
            gog_config = get_gog_config_from_path(self.interpreter.target_path)
            if gog_config:
                gog_game_path = get_gog_game_path(self.interpreter.target_path)
                lutris_config = convert_gog_config_to_lutris(gog_config, gog_game_path)
                self.script["game"].update(lutris_config)

        configpath = write_game_config(self.slug, self.get_game_config())
        self.game_id = add_or_update(
            name=self.game_name,
            runner=self.runner,
            slug=self.game_slug,
            platform=import_runner(self.runner)().get_platform(),
            directory=self.interpreter.target_path,
            installed=1,
            installer_slug=self.slug,
            parent_slug=self.requires,
            year=self.year,
            configpath=configpath,
            service=self.service.id if self.service else None,
            service_id=self.service_appid,
            id=self.game_id,
            discord_id=self.discord_id,
        )
        return self.game_id



import os

from lutris.util.log import logger


class RetroConfig:
    value_map = {"true": True, "false": False, "": None}

    def __init__(self, config_path):
        if not config_path:
            raise ValueError("Config path is mandatory")
        self.config_path = config_path
        self._config = []

    @property
    def config(self):
        """Lazy loading of the RetroArch config"""
        if self._config:
            return self._config
        try:
            self.load_config()
            return self._config
        except UnicodeDecodeError:
            logger.error(
                "The Retroarch config in %s could not be read because of character encoding issues", self.config_path
            )
            return []

    def load_config(self):
        """Load the configuration from file"""
        self._config = []
        if not os.path.isfile(self.config_path):
            raise OSError("Specified config file {} does not exist".format(self.config_path))
        with open(self.config_path, "r", encoding="utf-8") as config_file:
            for line in config_file.readlines():
                if not line:
                    continue
                line = line.strip()
                if line == "" or line.startswith("#"):
                    continue
                if "=" in line:
                    key, value = line.split("=", 1)
                    key = key.strip()
                    value = value.strip().strip('"')
                    if not key or not value:
                        continue
                    self._config.append((key, value))

    def save(self):
        with open(self.config_path, "w", encoding="utf-8") as config_file:
            for key, value in self.config:
                config_file.write('{} = "{}"\n'.format(key, value))

    def serialize_value(self, value):
        for k, v in self.value_map.items():
            if value is v:
                return k
        return value

    def deserialize_value(self, value):
        for k, v in self.value_map.items():
            if value == k:
                return v
        return value

    def __getitem__(self, key):
        for k, value in self.config:
            if key == k:
                return self.deserialize_value(value)

    def __setitem__(self, key, value):
        for index, conf in enumerate(self.config):
            if key == conf[0]:
                # self.config is read-only
                self._config[index] = (key, self.serialize_value(value))
                return
        self._config.append((key, self.serialize_value(value)))

    def keys(self):
        return [key for (key, _value) in self.config]



"""Utility module for creating an application wide logger."""

import logging
import logging.handlers
import os
import sys

from gi.repository import GLib

# Used to store log buffers for games.
LOG_BUFFERS = {}

CACHE_DIR = os.path.realpath(os.path.join(GLib.get_user_cache_dir(), "lutris"))
if not os.path.isdir(CACHE_DIR):
    os.makedirs(CACHE_DIR)

# Formatters
FILE_FORMATTER = logging.Formatter("[%(levelname)s:%(asctime)s:%(module)s]: %(message)s")
SIMPLE_FORMATTER = logging.Formatter("%(asctime)s: %(message)s")
DEBUG_FORMATTER = logging.Formatter("%(levelname)-8s %(asctime)s [%(module)s.%(funcName)s:%(lineno)s]:%(message)s")

# Log file setup
LOG_FILENAME = os.path.join(CACHE_DIR, "lutris.log")
file_handler = logging.handlers.RotatingFileHandler(LOG_FILENAME, backupCount=5)
file_handler.setFormatter(FILE_FORMATTER)

file_logger = logging.getLogger(__name__)
file_logger.setLevel(logging.DEBUG)
file_logger.addHandler(file_handler)

# Set the logging level to show debug messages.
console_handler = logging.StreamHandler(stream=sys.stderr)
console_handler.setFormatter(SIMPLE_FORMATTER)

logger = logging.getLogger(__name__)
logger.addHandler(console_handler)
logger.setLevel(logging.INFO)
logger.propagate = False


def get_log_contents() -> str:
    """Returns the entire text of the log file for this run."""
    if not os.path.exists(LOG_FILENAME):
        return ""
    with open(LOG_FILENAME, encoding="utf-8") as log_file:
        content = log_file.read()
    return content



"""Module that actually runs the games."""

# pylint: disable=too-many-public-methods disable=too-many-lines
import json
import os
import shlex
import shutil
import signal
import subprocess
import time
from gettext import gettext as _
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union, cast

from gi.repository import Gio, GLib, Gtk

from lutris import settings
from lutris.config import LutrisConfig
from lutris.database import categories as categories_db
from lutris.database import games as games_db
from lutris.database import sql
from lutris.exception_backstops import watch_game_errors
from lutris.exceptions import GameConfigError, InvalidGameMoveError, MissingExecutableError
from lutris.gui.widgets import NotificationSource
from lutris.installer import InstallationKind
from lutris.monitored_command import MonitoredCommand
from lutris.runner_interpreter import export_bash_script, get_launch_parameters
from lutris.runners import import_runner, is_valid_runner_name
from lutris.runners.runner import Runner, kill_processes
from lutris.util import busy, discord, extract, jobs, linux, strings, system, xdgshortcuts
from lutris.util.display import (
    DISPLAY_MANAGER,
    SCREEN_SAVER_INHIBITOR,
    disable_compositing,
    enable_compositing,
    is_display_x11,
)
from lutris.util.graphics.xrandr import turn_off_except
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import LOG_BUFFERS, logger
from lutris.util.steam.shortcut import remove_shortcut as remove_steam_shortcut
from lutris.util.system import fix_path_case
from lutris.util.timer import Timer
from lutris.util.yaml import write_yaml_to_file

if TYPE_CHECKING:
    from lutris.gui.application import LutrisApplication
    from lutris.gui.dialogs.delegates import LaunchUIDelegate

HEARTBEAT_DELAY = 2000

GAME_START = NotificationSource()
GAME_STARTED = NotificationSource()
GAME_STOPPED = NotificationSource()
GAME_UPDATED = NotificationSource()
GAME_INSTALLED = NotificationSource()
GAME_UNHANDLED_ERROR = NotificationSource()


class Game:
    """This class takes cares of loading the configuration for a game
    and running it.
    """

    now_playing_path = os.path.join(settings.CACHE_DIR, "now-playing.txt")

    STATE_STOPPED = "stopped"
    STATE_LAUNCHING = "launching"
    STATE_RUNNING = "running"

    PRIMARY_LAUNCH_CONFIG_NAME = "(primary)"

    def __init__(self, game_id: Optional[str] = None):
        super().__init__()

        self.game_error = NotificationSource()

        self._id = str(game_id) if game_id else None  # pylint: disable=invalid-name

        # Load attributes from database
        game_data = games_db.get_game_by_field(game_id, "id")

        self.slug = game_data.get("slug") or ""
        self._runner_name = game_data.get("runner") or ""
        self.directory = game_data.get("directory") or ""
        self.name = game_data.get("name") or ""
        self.sortname = game_data.get("sortname") or ""
        self.game_config_id = game_data.get("configpath") or ""
        self.is_installed = bool(game_data.get("installed") and self.game_config_id)
        self.platform = game_data.get("platform") or ""
        self.year = game_data.get("year") or ""
        self.lastplayed = game_data.get("lastplayed") or 0
        self.custom_images = set()
        if game_data.get("has_custom_banner"):
            self.custom_images.add("banner")
        if game_data.get("has_custom_icon"):
            self.custom_images.add("icon")
        if game_data.get("has_custom_coverart_big"):
            self.custom_images.add("coverart_big")
        self.service = game_data.get("service")
        self.appid = game_data.get("service_id")
        try:
            self.playtime = float(game_data.get("playtime") or 0.0)
        except ValueError as ex:
            logger.exception("Unable to parse playtime '%s' for game %s: %s", game_data.get("playtime"), self.slug, ex)
            self.playtime = 0.0

        self.discord_id = game_data.get("discord_id")  # Discord App ID for RPC

        self._config = None
        self._runner = None

        self.game_uuid = None
        self.game_thread = None
        self.antimicro_thread = None
        self.prelaunch_pids = None
        self.prelaunch_executor = None
        self.heartbeat = None
        self.killswitch = None
        self.state = self.STATE_STOPPED
        self.game_runtime_config = {}
        self.resolution_changed = False
        self.compositor_disabled = False
        self.original_outputs = None
        self._log_buffer = None
        self.timer = Timer()
        self.screen_saver_inhibitor_cookie = None

    @staticmethod
    def create_empty_service_game(db_game: Dict[str, Union[str, int]], service: Any) -> Any:
        """Creates a Game from the database data from ServiceGameCollection, which is
        not a real game, but which can be used to install. Such a game has no ID, but
        has an 'appid' and slug."""
        game = Game()
        game.name = db_game["name"]
        game.slug = service.get_installed_slug(db_game)
        game.runner_name = service.get_installed_runner_name(db_game)

        if "service_id" in db_game:
            game.appid = db_game["service_id"]
        elif service:
            game.appid = db_game["appid"]

        game.service = service.id if service else None
        return game

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        value = self.name or "Game (no name)"
        if self.runner_name:
            value += " (%s)" % self.runner_name
        return value

    @property
    def id(self) -> str:
        if not self._id:
            logger.error("The game '%s' has no ID, it is not stored in the database.", self.name)
        return cast(str, self._id)

    @property
    def is_db_stored(self) -> bool:
        """True if this Game has an ID, which means it is saved in the database."""
        return bool(self._id)

    @property
    def is_updatable(self) -> bool:
        """Return whether the game can be upgraded"""
        return self.is_installed and self.service in ["gog", "itchio"]

    def get_categories(self) -> List[Optional[str]]:
        """Return the categories the game is in."""
        return categories_db.get_categories_in_game(self.id) if self.is_db_stored else []

    def update_game_categories(self, added_category_names: list, removed_category_names: list) -> None:
        """add to / remove from categories"""
        for added_category_name in added_category_names:
            self.add_category(added_category_name, no_signal=True)

        for removed_category_name in removed_category_names:
            self.remove_category(removed_category_name, no_signal=True)

        GAME_UPDATED.fire(self)

    def add_category(self, category_name: str, no_signal=False) -> None:
        """add game to category"""
        if not self.is_db_stored:
            raise RuntimeError("Games that do not have IDs cannot belong to categories.")

        category = categories_db.get_category_by_name(category_name)
        if category is None:
            category_id = categories_db.add_category(category_name)
        else:
            category_id = category["id"]
        categories_db.add_game_to_category(self.id, category_id)

        if not no_signal:
            GAME_UPDATED.fire(self)

    def remove_category(self, category_name: str, no_signal=False) -> None:
        """remove game from category"""
        if not self.is_db_stored:
            return

        category = categories_db.get_category_by_name(category_name)
        if category is None:
            return
        category_id = category["id"]
        categories_db.remove_category_from_game(self.id, category_id)

        if not no_signal:
            GAME_UPDATED.fire(self)

    @property
    def is_favorite(self) -> bool:
        """Return whether the game is in the user's favorites"""
        return "favorite" in self.get_categories()

    def mark_as_favorite(self, is_favorite: bool) -> None:
        """Place the game in the favorite's category, or remove it.
        This change is applied at once, and does not need to be saved."""
        if self.is_favorite != bool(is_favorite):
            if is_favorite:
                self.add_category("favorite")
            else:
                self.remove_category("favorite")

    @property
    def is_hidden(self) -> bool:
        """Return whether the game is in the user's favorites"""
        return ".hidden" in self.get_categories()

    def mark_as_hidden(self, is_hidden: bool) -> None:
        """Place the game in the hidden category, or remove it.
        This change is applied at once, and does not need to be saved."""
        if self.is_hidden != bool(is_hidden):
            if is_hidden:
                self.add_category(".hidden")
            else:
                self.remove_category(".hidden")

    @property
    def log_buffer(self) -> Gtk.TextBuffer:
        """Access the log buffer object, creating it if necessary"""
        _log_buffer = LOG_BUFFERS.get(self.id)
        if _log_buffer:
            return _log_buffer
        _log_buffer = Gtk.TextBuffer()
        _log_buffer.create_tag("warning", foreground="red")
        if self.game_thread:
            self.game_thread.set_log_buffer(self._log_buffer)
            _log_buffer.set_text(self.game_thread.stdout)
        LOG_BUFFERS[self.id] = _log_buffer
        return _log_buffer

    @property
    def formatted_playtime(self) -> str:
        """Return a human-readable formatted play time"""
        return strings.get_formatted_playtime(self.playtime)

    def signal_error(self, error: Any) -> None:
        """Reports an error by firing game_error. If there is no handler for this,
        we fall back to the global GAME_UNHANDLED_ERROR.

        This allows special error handling to be set up for a particular Game, but
        there's always some global handling."""

        if self.game_error.has_handlers:
            self.game_error.fire(error)
        else:
            GAME_UNHANDLED_ERROR.fire(self, error)

    def get_browse_dir(self) -> str:
        """Return the path to open with the Browse Files action."""
        return self.resolve_game_path()

    def resolve_game_path(self) -> str:
        """Return the game's directory; if it is not known this will try to find
        it. This can still return an empty string if it can't do that."""
        if self.directory:
            return os.path.expanduser(self.directory)  # expanduser just in case!
        if self.has_runner:
            return self.runner.resolve_game_path()
        return ""

    def find_working_dir(self) -> str:
        """Returns a working directory- if we can't get one from the runner,
        we'll use the game path as a fallback."""
        if self.has_runner:
            return self.runner.working_dir or self.resolve_game_path()
        else:
            return self.resolve_game_path()

    @property
    def config(self) -> Union[LutrisConfig, None]:
        if not self.is_installed or not self.game_config_id:
            return None
        if not self._config:
            self._config = LutrisConfig(runner_slug=self.runner_name, game_config_id=self.game_config_id)
        return self._config

    @config.setter
    def config(self, value: LutrisConfig) -> None:
        self._config = value
        self._runner = None
        if value:
            self.game_config_id = value.game_config_id

    def reload_config(self) -> None:
        """Triggers the config to reload when next used; this also reloads the runner,
        so that it will pick up the new configuration."""
        self._config = None
        self._runner = None

    @property
    def runner_name(self) -> str:
        return self._runner_name

    @runner_name.setter
    def runner_name(self, value: str) -> None:
        self._runner_name = value or ""
        if self._runner and self._runner.name != value:
            self._runner = None

    @property
    def has_runner(self) -> bool:
        return bool(self._runner_name and is_valid_runner_name(self._runner_name))

    @property
    def runner(self) -> Runner:
        if not self.has_runner:
            raise GameConfigError(_("Invalid game configuration: Missing runner"))

        if not self._runner:
            runner_class = import_runner(self.runner_name)
            self._runner = runner_class(self.config)
        return cast(Runner, self._runner)

    @runner.setter
    def runner(self, value: Runner) -> None:
        self._runner = value
        if value:
            self._runner_name = value.name

    def set_desktop_compositing(self, enable: bool) -> None:
        """Enables or disables compositing"""
        if enable:
            if self.compositor_disabled:
                enable_compositing()
                self.compositor_disabled = False
        else:
            if not self.compositor_disabled:
                disable_compositing()
                self.compositor_disabled = True

    def install(self, launch_ui_delegate: "LaunchUIDelegate") -> None:
        """Request installation of a game"""
        if not self.slug:
            raise ValueError("Invalid game passed: %s" % self)

        if not self.service or self.service == "lutris":
            application: "LutrisApplication" = Gio.Application.get_default()
            application.show_lutris_installer_window(game_slug=self.slug)
            return

        service = launch_ui_delegate.get_service(self.service)
        db_game = service.get_service_db_game(self)
        if not db_game:
            logger.error("Can't find %s for %s, trying to fall back to Lutris installers", self.name, service.name)
            application: "LutrisApplication" = Gio.Application.get_default()
            application.show_lutris_installer_window(game_slug=self.slug)
            return

        try:
            game_id = service.install(db_game)
        except ValueError as e:
            logger.debug(e)
            game_id = None

        if game_id:

            def on_error(error: BaseException) -> None:
                logger.exception("Unable to install game: %s", error)

            game = Game(game_id)
            game.game_error.register(on_error)
            game.launch(launch_ui_delegate)

    def install_updates(self, install_ui_delegate: "LaunchUIDelegate") -> bool:
        service = install_ui_delegate.get_service(self.service)
        db_game = games_db.get_game_by_field(self.id, "id")

        def on_installers_ready(installers, error):
            if error:
                raise error  # bounce errors off the backstop

            if not installers:
                raise RuntimeError(_("No updates found"))

            application = Gio.Application.get_default()
            application.show_installer_window(
                installers, service, self.appid, installation_kind=InstallationKind.UPDATE
            )

        busy.BusyAsyncCall(service.get_update_installers, on_installers_ready, db_game)
        return True

    def install_dlc(self, install_ui_delegate: "LaunchUIDelegate") -> bool:
        service = install_ui_delegate.get_service(self.service)
        db_game = games_db.get_game_by_field(self.id, "id")

        def on_installers_ready(installers, error):
            if error:
                raise error  # bounce errors off the backstop

            if not installers:
                raise RuntimeError(_("No DLC found"))

            application = Gio.Application.get_default()
            application.show_installer_window(installers, service, self.appid, installation_kind=InstallationKind.DLC)

        busy.BusyAsyncCall(service.get_dlc_installers_runner, on_installers_ready, db_game, db_game["runner"])
        return True

    def uninstall(self, delete_files: bool = False) -> None:
        """Uninstall a game, but do not remove it from the library.

        Params:
            delete_files (bool): Delete the game files
        """
        sql.db_update(settings.DB_PATH, "games", {"installed": 0, "runner": ""}, {"id": self.id})
        if self.config:
            self.config.remove()
        xdgshortcuts.remove_launcher(self.slug, self.id, desktop=True, menu=True)
        remove_steam_shortcut(self)
        if delete_files and self.has_runner:
            # self.directory here, not self.resolve_game_path; no guessing at
            # directories when we delete them
            self.runner.remove_game_data(app_id=self.appid, game_path=self.directory)
        self.is_installed = False
        self._config = None
        self._runner = None

        if self.id in LOG_BUFFERS:  # Reset game logs on removal
            log_buffer = LOG_BUFFERS[self.id]
            log_buffer.delete(log_buffer.get_start_iter(), log_buffer.get_end_iter())

    def delete(self) -> None:
        """Delete a game from the library; must be uninstalled first."""
        if self.is_installed:
            raise RuntimeError(_("Uninstall the game before deleting"))
        games_db.delete_game(self.id)
        self._id = None

    def set_platform_from_runner(self) -> None:
        """Set the game's platform from the runner"""
        if not hasattr(self, "has_runner") or not self.has_runner:
            logger.warning("Game has no runner, can't set platform")
            return
        self.platform = self.runner.get_platform()
        if not self.platform:
            logger.warning("The %s runner didn't provide a platform for %s", self.runner, self)

    def save(self, no_signal: bool = False) -> None:
        """
        Save the game's config and metadata.
        """
        if self.config:
            configpath = self.config.game_config_id
            logger.debug("Saving %s with config ID %s", self, self.config.game_config_id)
            self.config.save()
        else:
            logger.warning("Saving %s with the configuration missing", self)
            configpath = ""
        self.set_platform_from_runner()

        game_data = {
            "name": self.name,
            "sortname": self.sortname,
            "runner": self.runner_name,
            "slug": self.slug,
            "platform": self.platform,
            "directory": self.directory,
            "installed": self.is_installed,
            "year": self.year,
            "lastplayed": self.lastplayed,
            "configpath": configpath,
            "id": self.id,
            "playtime": self.playtime,
            "service": self.service,
            "service_id": self.appid,
            "discord_id": self.discord_id,
            "has_custom_banner": "banner" in self.custom_images,
            "has_custom_icon": "icon" in self.custom_images,
            "has_custom_coverart_big": "coverart_big" in self.custom_images,
        }
        self._id = str(games_db.add_or_update(**game_data))
        if not no_signal:
            GAME_UPDATED.fire(self)

    def save_platform(self) -> None:
        """Save only the platform field- do not restore any other values the user may have changed
        in another window."""
        games_db.update_existing(id=self.id, slug=self.slug, platform=self.platform)
        GAME_UPDATED.fire(self)

    def save_lastplayed(self) -> None:
        """Save only the lastplayed field- do not restore any other values the user may have changed
        in another window."""
        games_db.update_existing(id=self.id, slug=self.slug, lastplayed=self.lastplayed, playtime=self.playtime)
        GAME_UPDATED.fire(self)

    def check_launchable(self) -> bool:
        """Verify that the current game can be launched, and raises exceptions if not."""
        if not self.is_installed or not self.is_db_stored:
            logger.error("%s (%s) not installed", self, self.id)
            raise GameConfigError(_("Tried to launch a game that isn't installed."))
        if not self.has_runner:
            raise GameConfigError(_("Invalid game configuration: Missing runner"))

        return True

    def restrict_to_display(self, display: str) -> bool:
        outputs = DISPLAY_MANAGER.get_config()
        if display == "primary":
            display = ""
            for output in outputs:
                if output.primary:
                    display = output.name
                    break
            if not display:
                logger.warning("No primary display set")
        else:
            found = False
            for output in outputs:
                if output.name == display:
                    found = True
                    break
            if not found:
                logger.warning("Selected display %s not found", display)
                display = ""
        if display:
            turn_off_except(display)
            time.sleep(3)
            return True
        return False

    def start_antimicrox(self, antimicro_config) -> None:
        """Start Antimicrox with a given config path"""
        if LINUX_SYSTEM.is_flatpak():
            antimicro_command = ["flatpak-spawn", "--host", "antimicrox"]
        else:
            try:
                antimicro_command = [system.find_required_executable("antimicrox")]
            except MissingExecutableError as ex:
                raise GameConfigError(
                    _("Unable to find Antimicrox, install it or disable the Antimicrox option")
                ) from ex

        logger.info("Starting Antimicro")
        antimicro_command += ["--hidden", "--tray", "--profile", antimicro_config]
        self.antimicro_thread = MonitoredCommand(antimicro_command)
        if self.antimicro_thread and hasattr(self.antimicro_thread, "start"):
            self.antimicro_thread.start()

    def start_prelaunch_command(self, wait_for_completion=False) -> None:
        """Start the prelaunch command specified in the system options"""
        prelaunch_command = self.runner.system_config.get("prelaunch_command")
        if not prelaunch_command:
            return
        command_array = shlex.split(prelaunch_command)
        if not system.path_exists(command_array[0]):
            logger.warning("Command %s not found", command_array[0])
            return
        env = self.game_runtime_config["env"]
        cwd = self.find_working_dir()
        if wait_for_completion:
            logger.info("Prelauch command: %s, waiting for completion", prelaunch_command)
            # Monitor the prelaunch command and wait until it has finished
            system.execute(command_array, env=env, cwd=cwd)
        else:
            logger.info("Prelaunch command %s launched in the background", prelaunch_command)
            self.prelaunch_executor = MonitoredCommand(
                command_array,
                include_processes=[os.path.basename(command_array[0])],
                env=env,
                cwd=cwd,
            )
            if not hasattr(self.prelaunch_executor, "start") or not self.prelaunch_executor:
                raise RuntimeError("Prelaunch Executor: %s doesn't have a start attribute" % self.prelaunch_executor)
            self.prelaunch_executor.start()

    def get_terminal(self) -> str:
        """Return the terminal used to run the game into or None if the game is not run from a terminal.
        Remember that only games using text mode should use the terminal.
        """
        if self.runner.system_config.get("terminal"):
            terminal = self.runner.system_config.get("terminal_app", linux.get_default_terminal())
            if terminal and not system.can_find_executable(terminal):
                raise GameConfigError(_("The selected terminal application could not be launched:\n%s") % terminal)
            return terminal or ""
        return ""

    def get_killswitch(self) -> str:
        """Return the path to a file that is monitored during game execution.
        If the file stops existing, the game is stopped.
        """
        killswitch = self.runner.system_config.get("killswitch")
        # Prevent setting a killswitch to a file that doesn't exists
        if killswitch and system.path_exists(killswitch):
            return killswitch
        return ""

    def get_gameplay_info(self, launch_ui_delegate: "LaunchUIDelegate"):
        """Return the information provided by a runner's play method.
        It checks for possible errors and raises exceptions if they occur.

        This may invoke methods on the delegates to make decisions,
        and this may show UI.

        This returns an empty dictionary if the user cancels this UI,
        in which case the game should not be run.
        """
        if not hasattr(self.runner, "play") or not callable(getattr(self.runner, "play", None)):
            raise RuntimeError(f"Runner: {self.runner} doesn't have a play method.")
        gameplay_info = self.runner.play()  # type: ignore[attr-defined]

        if "working_dir" not in gameplay_info:
            gameplay_info["working_dir"] = self.runner.working_dir

        config = launch_ui_delegate.select_game_launch_config(self)

        if config is None:
            return {}  # no error here- the user cancelled out

        if config:  # empty dict for primary configuration
            self.runner.apply_launch_config(gameplay_info, config)

        return gameplay_info

    def get_path_from_config(self) -> str:
        """Return the path of the main entry point for a game"""
        if not self.config:
            logger.warning("%s has no configuration", self)
            return ""
        game_config = self.config.game_config

        # Skip MAME roms referenced by their ID
        if self.runner_name == "mame":
            if "main_file" in game_config and "." not in game_config["main_file"]:
                return ""

        for key in ["exe", "main_file", "iso", "rom", "disk-a", "path", "files"]:
            if key in game_config:
                path = game_config[key]
                if key == "files":
                    path = path[0]

                if path:
                    path = os.path.expanduser(path)
                    if not path.startswith("/"):
                        path = os.path.join(self.directory, path)

                    # The Wine runner fixes case mismatches automatically,
                    # sort of like Windows, so we need to do the same.
                    if self.runner_name == "wine":
                        path = fix_path_case(path)

                    return path

        logger.warning("No path found in %s", self.config)
        return ""

    def get_store_name(self) -> str:
        store = self.service
        if not store:
            return ""
        if self.service == "humblebundle":
            return "humble"
        return store

    @watch_game_errors(game_stop_result=False)
    def configure_game(self, launch_ui_delegate: "LaunchUIDelegate") -> bool:
        """Get the game ready to start, applying all the options.
        This method sets the game_runtime_config attribute.
        """
        gameplay_info = self.get_gameplay_info(launch_ui_delegate)
        if not gameplay_info:  # if user cancelled - not an error
            return False
        command, env = get_launch_parameters(self.runner, gameplay_info)

        store = env.get("STORE") or self.get_store_name()
        if store:
            env["STORE"] = store

        # Some environment variables for the use of custom pre-launch and post-exit scripts.
        env["GAME_NAME"] = self.name
        if self.directory:
            env["GAME_DIRECTORY"] = self.directory

        self.game_runtime_config = {
            "args": command,
            "env": env,
            "terminal": self.get_terminal(),
            "include_processes": shlex.split(self.runner.system_config.get("include_processes", "")),
            "exclude_processes": shlex.split(self.runner.system_config.get("exclude_processes", "")),
        }

        if "working_dir" in gameplay_info:
            self.game_runtime_config["working_dir"] = gameplay_info["working_dir"]

        # Input control
        if self.runner.system_config.get("use_us_layout") and is_display_x11():
            system.set_keyboard_layout("us")

        # Display control
        self.original_outputs = DISPLAY_MANAGER.get_config()

        if self.runner.system_config.get("disable_compositor"):
            self.set_desktop_compositing(False)

        if self.runner.system_config.get("disable_screen_saver"):
            self.screen_saver_inhibitor_cookie = SCREEN_SAVER_INHIBITOR.inhibit(self.name)

        if self.runner.system_config.get("display") != "off":
            self.resolution_changed = self.restrict_to_display(self.runner.system_config.get("display") or "")

        resolution = self.runner.system_config.get("resolution")
        if resolution != "off":
            DISPLAY_MANAGER.set_resolution(resolution)
            time.sleep(3)
            self.resolution_changed = True

        self.runner.finish_env(env, self)

        antimicro_config = self.runner.system_config.get("antimicro_config", "")
        if system.path_exists(antimicro_config):
            self.start_antimicrox(antimicro_config)

        # Execution control
        self.killswitch = self.get_killswitch()

        if self.runner.system_config.get("prelaunch_command", ""):
            self.start_prelaunch_command(self.runner.system_config["prelaunch_wait"])

        self.start_game()
        return True

    @watch_game_errors(game_stop_result=False)
    def launch(self, launch_ui_delegate) -> bool:
        """Request launching a game. The game may not be installed yet."""
        if not self.check_launchable():
            logger.error("Game is not launchable")
            return False

        if not launch_ui_delegate.check_game_launchable(self):
            return False

        self.reload_config()  # Reload the config before launching it.

        if self.id in LOG_BUFFERS:  # Reset game logs on each launch
            log_buffer = LOG_BUFFERS[self.id]
            log_buffer.delete(log_buffer.get_start_iter(), log_buffer.get_end_iter())

        self.state = self.STATE_LAUNCHING
        self.prelaunch_pids = system.get_running_pid_list()

        if not self.prelaunch_pids:
            logger.error("No prelaunch PIDs could be obtained. Game stop may be ineffective.")
            self.prelaunch_pids = None

        GAME_START.fire(self)

        @watch_game_errors(game_stop_result=False, game=self)
        def configure_game(_ignored, error) -> None:
            if error:
                raise error
            self.configure_game(launch_ui_delegate)

        jobs.AsyncCall(self.runner.prelaunch, configure_game)
        return True

    def start_game(self) -> None:
        """Run a background command to launch the game"""
        self.game_thread = MonitoredCommand(
            self.game_runtime_config["args"],
            title=self.name,
            runner=self.runner,
            cwd=self.game_runtime_config.get("working_dir"),
            env=self.game_runtime_config["env"],
            term=self.game_runtime_config["terminal"],
            log_buffer=self.log_buffer,
            include_processes=self.game_runtime_config["include_processes"],
            exclude_processes=self.game_runtime_config["exclude_processes"],
        )
        stop_func = getattr(self.runner, "stop", None)
        if stop_func and self.game_thread:
            self.game_thread.stop_func = stop_func

        if self.game_thread:
            self.game_uuid = self.game_thread.env["LUTRIS_GAME_UUID"]
            self.game_thread.start()

        self.timer.start()
        self.state = self.STATE_RUNNING
        GAME_STARTED.fire(self)

        # Game is running, let's update discord status
        if settings.read_setting("discord_rpc") == "True" and self.discord_id:
            try:
                discord.client.update(self.discord_id)
            except AssertionError:
                pass

        self.heartbeat = GLib.timeout_add(HEARTBEAT_DELAY, self.beat)
        with open(self.now_playing_path, "w", encoding="utf-8") as np_file:
            np_file.write(self.name)

    def force_stop(self) -> None:
        # If force_stop_game fails, wait a few seconds and try SIGKILL on any survivors

        def force_stop_game() -> bool:
            self.runner.force_stop_game(self.get_stop_pids())
            return not self.get_stop_pids()

        def force_stop_game_cb(all_dead, error) -> None:
            if error:
                self.signal_error(error)
            elif all_dead:
                self.stop_game()
            else:
                self.force_kill_delayed()

        busy.BusyAsyncCall(force_stop_game, force_stop_game_cb)

    def force_kill_delayed(self, death_watch_seconds: int = 5, death_watch_interval_seconds: float = 0.5) -> None:
        """Forces termination of a running game, but only after a set time has elapsed;
        Invokes stop_game() when the game is dead."""

        def death_watch() -> None:
            """Wait for the processes to die; kill them if they do not."""
            for _n in range(int(death_watch_seconds / death_watch_interval_seconds)):
                time.sleep(death_watch_interval_seconds)
                if not self.get_stop_pids():
                    return

            # Once we get past the time limit, starting killing!
            kill_processes(signal.SIGKILL, self.get_stop_pids())

        def death_watch_cb(_result, error) -> None:
            """Called after the death watch to more firmly kill any survivors."""
            if error:
                self.signal_error(error)

            # If we still can't kill everything, we'll still say we stopped it.
            self.stop_game()

        busy.BusyAsyncCall(death_watch, death_watch_cb)

    def get_stop_pids(self) -> set:
        """Finds the PIDs of processes that need killin'!"""
        pids = self.get_game_pids()
        if self.game_thread and self.game_thread.game_process:
            if self.game_thread.game_process.poll() is None:
                pids.add(self.game_thread.game_process.pid)
        return pids

    def get_game_pids(self) -> set:
        """Return a list of processes belonging to the Lutris game"""
        if not self.game_uuid:
            logger.error("No LUTRIS_GAME_UUID recorded. The game's PIDs cannot be computed.")
            return set()

        new_pids = self.get_new_pids()
        game_folder = self.resolve_game_path()
        return self.runner.filter_game_pids(new_pids, self.game_uuid, game_folder)

    def get_new_pids(self) -> set:
        """Return list of PIDs started since the game was launched"""
        if self.prelaunch_pids:
            return set(system.get_running_pid_list()) - set(self.prelaunch_pids)

        logger.error("No prelaunch PIDs recorded. The game's PIDs cannot be computed.")
        return set()

    def stop_game(self) -> None:
        """Cleanup after a game as stopped"""
        duration = self.timer.duration
        logger.debug("%s has run for %d seconds", self, duration)
        if duration < 5:
            logger.warning("The game has run for a very short time, did it crash?")
            # Inspect why it could have crashed

        self.state = self.STATE_STOPPED
        GAME_STOPPED.fire(self)
        if os.path.exists(self.now_playing_path):
            os.unlink(self.now_playing_path)
        if not self.timer.finished:
            self.timer.end()
            self.playtime += self.timer.duration / 3600
            logger.debug("Playtime: %s", self.formatted_playtime)
            self.save_lastplayed()

    @watch_game_errors(game_stop_result=False)
    def beat(self) -> bool:
        """Watch the game's process(es)."""
        if not self.game_thread:
            return True
        if self.game_thread.error:
            self.on_game_quit()
            raise RuntimeError(_("<b>Error lauching the game:</b>\n") + self.game_thread.error)

        # The killswitch file should be set to a device (ie. /dev/input/js0)
        # When that device is unplugged, the game is forced to quit.
        killswitch_engage = self.killswitch and not system.path_exists(self.killswitch)
        if killswitch_engage:
            logger.warning("File descriptor no longer present, force quit the game")
            self.force_stop()
            return False
        game_pids = self.get_game_pids()
        runs_only_prelaunch = False
        if self.prelaunch_executor and self.prelaunch_executor.is_running and self.prelaunch_executor.game_process:
            runs_only_prelaunch = game_pids == {self.prelaunch_executor.game_process.pid}
        if runs_only_prelaunch or (not self.game_thread.is_running and not game_pids):
            logger.debug("Game thread stopped")
            self.on_game_quit()
            return False

        # If game thread stopped but orphan processes remain (e.g. gamescope), clean them up
        if not self.game_thread.is_running and game_pids:
            logger.debug("Game thread stopped but %d orphan processes remain, cleaning up", len(game_pids))
            self.runner.force_stop_game(game_pids)
            self.on_game_quit()
            return False

        return True

    def stop(self) -> None:
        """Stops the game"""
        if self.state == self.STATE_STOPPED:
            logger.debug("Game already stopped")
            return

        logger.info("Stopping %s", self)

        if self.game_thread:

            def stop_cb(_result, error):
                if error:
                    self.signal_error(error)

            jobs.AsyncCall(self.game_thread.stop, stop_cb)
        self.stop_game()

    def on_game_quit(self) -> None:
        """Restore some settings and cleanup after game quit."""

        if self.prelaunch_executor and self.prelaunch_executor.is_running:
            logger.info("Stopping prelaunch script")
            self.prelaunch_executor.stop()

        # We need to do some cleanup before we emit game-stop as this can
        # trigger Lutris shutdown

        if self.screen_saver_inhibitor_cookie is not None:
            SCREEN_SAVER_INHIBITOR.uninhibit(self.screen_saver_inhibitor_cookie)
            self.screen_saver_inhibitor_cookie = None

        self.heartbeat = None
        if self.state != self.STATE_STOPPED:
            logger.warning("Game still running (state: %s)", self.state)
            self.stop()

        # Check for post game script
        postexit_command = self.runner.system_config.get("postexit_command")
        if postexit_command:
            command_array = shlex.split(postexit_command)
            if system.path_exists(command_array[0]):
                logger.info("Running post-exit command: %s", postexit_command)
                postexit_thread = MonitoredCommand(
                    command_array,
                    include_processes=[os.path.basename(postexit_command)],
                    env=self.game_runtime_config["env"],
                    cwd=self.find_working_dir(),
                )
                postexit_thread.start()

        quit_time = time.strftime("%a, %d %b %Y %H:%M:%S", time.localtime())
        logger.debug("%s stopped at %s", self.name, quit_time)
        self.lastplayed = int(time.time())
        self.save_lastplayed()

        os.chdir(os.path.expanduser("~"))

        if self.antimicro_thread:
            self.antimicro_thread.stop()

        if self.resolution_changed or self.runner.system_config.get("reset_desktop"):
            DISPLAY_MANAGER.set_resolution(self.original_outputs)

        if self.compositor_disabled:
            self.set_desktop_compositing(True)

        if self.runner.system_config.get("use_us_layout"):
            with subprocess.Popen(["setxkbmap"], env=os.environ) as setxkbmap:
                setxkbmap.communicate()

        # Clear Discord Client Status
        if settings.read_setting("discord_rpc") == "True" and self.discord_id:
            try:
                discord.client.clear()
            except:
                # Shut up no one cares about you or your errors
                pass

        self.process_return_codes()

    def process_return_codes(self) -> None:
        """Do things depending on how the game quitted."""
        if not self.game_thread:
            return
        if self.game_thread.return_code == 127:
            # Error missing shared lib
            error = "error while loading shared lib"
            error_lines = strings.lookup_strings_in_text(error, self.game_thread.stdout)
            if error_lines:
                raise RuntimeError(_("<b>Error: Missing shared library.</b>\n\n%s") % error_lines[0])

        if self.game_thread.return_code == 1:
            # Error Wine version conflict
            error = "maybe the wrong wineserver"
            if strings.lookup_strings_in_text(error, self.game_thread.stdout):
                raise RuntimeError(_("<b>Error: A different Wine version is already using the same Wine prefix.</b>"))

    def write_script(self, script_path, launch_ui_delegate) -> None:
        """Output the launch argument in a bash script"""
        gameplay_info = self.get_gameplay_info(launch_ui_delegate)
        if not gameplay_info:
            # User cancelled; errors are raised as exceptions instead of this
            return
        export_bash_script(self.runner, gameplay_info, script_path)

    def move(self, new_location, no_signal=False) -> str:
        logger.info("Moving %s to %s", self, new_location)
        new_config = ""
        old_location = self.directory
        target_directory = self._get_move_target_directory(new_location)

        if old_location and system.path_contains(old_location, new_location):
            raise InvalidGameMoveError(
                _("Lutris can't move '%s' to a location inside of itself, '%s'.") % (old_location, new_location)
            )

        self.directory = target_directory
        self.save(no_signal=no_signal)

        if not old_location:
            # We can't move or update the config without an initial
            # location, but no-one expects us to. We've just updated
            # the game directory, and that will do.
            return target_directory

        if not self.config:
            raise RuntimeError(f"Game configuration not found for {self.name}.")

        config_path = self.config.game_config_path
        if not config_path:
            raise RuntimeError(f"Game configuration path not found for {self.name}.")

        with open(config_path, encoding="utf-8") as config_file:
            for line in config_file.readlines():
                if target_directory in line:
                    new_config += line
                else:
                    new_config += line.replace(old_location, target_directory)
        with open(config_path, "w", encoding="utf-8") as config_file:
            config_file.write(new_config)

        if not system.path_exists(old_location):
            logger.warning("Initial location %s does not exist, files may have already been moved.")
            return target_directory

        try:
            shutil.move(old_location, new_location)
        except OSError as ex:
            logger.error(
                "Failed to move %s to %s, you may have to move files manually (Exception: %s)",
                old_location,
                new_location,
                ex,
            )
        return target_directory

    def set_location(self, new_location) -> str:
        target_directory = self._get_move_target_directory(new_location)
        self.directory = target_directory
        self.save()
        return target_directory

    def _get_move_target_directory(self, new_location: str) -> str:
        old_location = self.directory
        if old_location and os.path.exists(old_location):
            game_directory = os.path.basename(old_location)
            return os.path.join(new_location, game_directory)

        return new_location


def export_game(slug: str, dest_dir: str) -> None:
    """Export a full game folder along with some lutris metadata"""
    # List of runner where we know for sure that 1 folder = 1 game.
    # For runners that handle ROMs, we have to handle this more finely.
    # There is likely more than one game in a ROM folder but a ROM
    # might have several files (like a bin/cue, or a multi-disk game)
    exportable_runners = [
        "linux",
        "wine",
        "dosbox",
        "scummvm",
    ]
    db_game = games_db.get_game_by_field(slug, "slug")
    if not db_game:
        logger.error("Game %s not found", slug)
        return
    if db_game["runner"] not in exportable_runners:
        raise RuntimeError("Game %s can't be exported." % db_game["name"])
    if not db_game["directory"]:
        raise RuntimeError("No game directory set. Could we guess it?")

    game = Game(db_game["id"])
    if game.config:
        db_game["config"] = game.config.game_level
    else:
        db_game["config"] = None
    game_path = db_game["directory"]
    config_path = os.path.join(db_game["directory"], "%s.lutris" % slug)
    with open(config_path, "w", encoding="utf-8") as config_file:
        json.dump(db_game, config_file, indent=2)
    archive_path = os.path.join(dest_dir, "%s.tar.xz" % slug)
    command = ["tar", "cJf", archive_path, os.path.basename(game_path)]
    system.execute(command, cwd=os.path.dirname(game_path))
    logger.info("%s exported to %s", slug, archive_path)


def import_game(file_path, dest_dir) -> None:
    """Import a game in Lutris"""
    if not os.path.exists(file_path):
        raise RuntimeError("No file %s" % file_path)
    logger.info("Importing %s to %s", file_path, dest_dir)
    if not os.path.isdir(dest_dir):
        os.makedirs(dest_dir)
    original_file_list = set(os.listdir(dest_dir))
    extract.extract_archive(file_path, dest_dir, merge_single=False)
    new_file_list = set(os.listdir(dest_dir))
    new_dir = list(new_file_list - original_file_list)[0]
    game_dir = os.path.join(dest_dir, new_dir)
    try:
        game_config = [f for f in os.listdir(game_dir) if f.endswith(".lutris")][0]
    except IndexError:
        logger.error("No Lutris configuration file found in archive")
        return

    with open(os.path.join(game_dir, game_config), encoding="utf-8") as config_file:
        lutris_config = json.load(config_file)
    old_dir = lutris_config["directory"]
    with open(os.path.join(game_dir, game_config), "r", encoding="utf-8") as config_file:
        config_data = config_file.read()
    config_data = config_data.replace(old_dir, game_dir)
    with open(os.path.join(game_dir, game_config), "w", encoding="utf-8") as config_file:
        config_file.write(config_data)
    with open(os.path.join(game_dir, game_config), encoding="utf-8") as config_file:
        lutris_config = json.load(config_file)
    config_filename = os.path.join(settings.CONFIG_DIR, "games/%s.yml" % lutris_config["configpath"])
    write_yaml_to_file(lutris_config["config"], config_filename)
    game_id = games_db.add_game(
        name=lutris_config["name"],
        runner=lutris_config["runner"],
        slug=lutris_config["slug"],
        platform=lutris_config["platform"],
        directory=game_dir,
        installed=lutris_config["installed"],
        year=lutris_config["year"],
        lastplayed=lutris_config["lastplayed"],
        configpath=lutris_config["configpath"],
        playtime=lutris_config["playtime"],
        service=lutris_config["service"],
        service_id=lutris_config["service_id"],
    )
    print("Added game with ID %s" % game_id)



from lutris.util.linux import LINUX_SYSTEM
from lutris.util.wine.dll_manager import DLLManager


class VKD3DManager(DLLManager):
    name = "vkd3d"
    human_name = "VKD3D"
    managed_dlls = ("d3d12", "d3d12core")
    releases_url = "https://api.github.com/repos/lutris/vkd3d/releases"

    def can_enable(self):
        return LINUX_SYSTEM.is_vulkan_supported()



"""Keep track of game executables' presence"""

import json
import os
import time

from lutris import settings
from lutris.database.games import get_games
from lutris.game import Game
from lutris.gui.widgets import NotificationSource
from lutris.util import cache_single
from lutris.util.jobs import AsyncCall
from lutris.util.log import logger

GAME_PATH_CACHE_PATH = os.path.join(settings.CACHE_DIR, "game-paths.json")


def get_game_paths():
    game_paths = {}
    all_games = get_games(filters={"installed": 1})
    for db_game in all_games:
        if db_game.get("runner") in ("steam", "web"):
            continue
        game = Game(db_game["id"])
        path = game.get_path_from_config()
        if not path:
            continue
        game_paths[db_game["id"]] = path
    return game_paths


def build_path_cache(recreate=False):
    """Generate a new cache path"""
    if os.path.exists(GAME_PATH_CACHE_PATH) and not recreate:
        return
    start_time = time.time()
    with open(GAME_PATH_CACHE_PATH, "w", encoding="utf-8") as cache_file:
        game_paths = get_game_paths()
        json.dump(game_paths, cache_file, indent=2)
    end_time = time.time()
    get_path_cache.cache_clear()
    logger.debug("Game path cache built in %0.2f seconds", end_time - start_time)


def add_to_path_cache(game):
    """Add or update the path of a game in the cache"""
    logger.debug("Adding %s to path cache", game)
    path = game.get_path_from_config()
    if not path:
        logger.warning("No path for %s", game)
        return
    current_cache = read_path_cache()
    current_cache[game.id] = path
    with open(GAME_PATH_CACHE_PATH, "w", encoding="utf-8") as cache_file:
        json.dump(current_cache, cache_file, indent=2)
    get_path_cache.cache_clear()


@cache_single
def get_path_cache():
    """Return the contents of the path cache file; this
    dict is cached, so do not modify it."""
    return read_path_cache()


def read_path_cache():
    """Read the contents of the path cache file, and does not cache it."""
    with open(GAME_PATH_CACHE_PATH, encoding="utf-8") as cache_file:
        try:
            return json.load(cache_file)
        except json.JSONDecodeError:
            return {}


def remove_from_path_cache(game):
    logger.debug("Removing %s from path cache", game)
    current_cache = read_path_cache()
    if game.id not in current_cache:
        logger.warning("Game %s (id=%s) not in cache path", game, game.id)
        return
    del current_cache[game.id]
    with open(GAME_PATH_CACHE_PATH, "w", encoding="utf-8") as cache_file:
        json.dump(current_cache, cache_file, indent=2)
    get_path_cache.cache_clear()


class MissingGames:
    """This class is a singleton that holds a set of game-ids for games whose directories
    are missing. It is updated on a background thread, but there's a NotificationSource ('updated')
    that fires when that thread has made changes and exited, so that the UI cab update then."""

    def __init__(self):
        self.updated = NotificationSource()
        self.missing_game_ids = set()
        self._update_running = None

    @property
    def is_initialized(self):
        """True if the missing games have ever been updated."""
        return self._update_running is not None

    def update_all_missing(self) -> None:
        """This starts the check for all games; the actual list of game-ids will be obtained
        on the worker thread, and this method will start it."""

        if not self._update_running:
            self._update_running = True
            AsyncCall(self._update_missing_games, self._update_missing_games_cb)

    def _update_missing_games(self):
        """This is the method that runs on the worker thread; it checks each game given
        and returns True if any changes to missing_game_ids was made."""

        logger.debug("Checking for missing games")

        changed = False

        for game_id, path in get_path_cache().items():
            if path:
                old_status = game_id in self.missing_game_ids
                new_status = not os.path.exists(path)
                if old_status != new_status:
                    if new_status:
                        self.missing_game_ids.add(game_id)
                    else:
                        self.missing_game_ids.discard(game_id)
                    changed = True
        return changed

    def _update_missing_games_cb(self, changed, error):
        self._update_running = False

        if error:
            logger.exception("Unable to detect missing games: %s", error)
        elif changed:
            self.updated.fire()


MISSING_GAMES = MissingGames()



"""Runtime handling module"""

import concurrent.futures
import os
import threading
import time
from gettext import gettext as _
from typing import Any, Dict, List, Optional

from lutris import settings
from lutris.api import (
    check_stale_runtime_versions,
    download_runtime_versions,
    get_runtime_versions,
    get_time_from_api_date,
)
from lutris.gui.widgets.progress_box import ProgressInfo
from lutris.util import http, system
from lutris.util.downloader import Downloader
from lutris.util.extract import extract_archive
from lutris.util.jobs import AsyncCall
from lutris.util.linux import LINUX_SYSTEM
from lutris.util.log import logger
from lutris.util.strings import parse_version
from lutris.util.wine.d3d_extras import D3DExtrasManager
from lutris.util.wine.dgvoodoo2 import dgvoodoo2Manager
from lutris.util.wine.dxvk import DXVKManager
from lutris.util.wine.dxvk_nvapi import DXVKNVAPIManager
from lutris.util.wine.vkd3d import VKD3DManager

RUNTIME_DISABLED = os.environ.get("LUTRIS_RUNTIME", "").casefold() in ("0", "off")
DEFAULT_RUNTIME = "Ubuntu-18.04"

DLL_MANAGERS = {
    "dxvk": DXVKManager,
    "vkd3d": VKD3DManager,
    "d3d_extras": D3DExtrasManager,
    "dgvoodoo2": dgvoodoo2Manager,
    "dxvk_nvapi": DXVKNVAPIManager,
}


def get_env(
    version: Optional[str] = None, prefer_system_libs: bool = False, wine_path: Optional[str] = None
) -> Dict[str, str]:
    """Return a dict containing LD_LIBRARY_PATH env var

    Params:
        version: Version of the runtime to use, such as "Ubuntu-18.04" or "legacy"
        prefer_system_libs: Whether to prioritize system libs over runtime libs
        wine_path: If you prioritize system libs, provide the path for a lutris wine build
                         if one is being used. This allows Lutris to prioritize the wine libs
                         over everything else.
    """
    library_path = ":".join(get_paths(version=version, prefer_system_libs=prefer_system_libs, wine_path=wine_path))
    env = {}
    if library_path:
        env["LD_LIBRARY_PATH"] = library_path
        network_tools_path = os.path.join(settings.RUNTIME_DIR, "network-tools")
        env["PATH"] = "%s:%s" % (network_tools_path, os.environ["PATH"])
    return env


def get_winelib_paths(wine_path: str) -> List[str]:
    """Return wine libraries path for a Lutris wine build"""
    paths = []
    # Prioritize libwine.so.1 for lutris builds
    for winelib_path in ("lib", "lib64"):
        winelib_fullpath = os.path.join(wine_path or "", winelib_path)
        if system.path_exists(winelib_fullpath):
            paths.append(winelib_fullpath)
    return paths


def get_runtime_paths(
    version: Optional[str] = None, prefer_system_libs: bool = True, wine_path: Optional[str] = None
) -> List[str]:
    """Return Lutris runtime paths"""
    version = version or DEFAULT_RUNTIME
    lutris_runtime_path = "%s-i686" % version
    runtime_paths = [
        lutris_runtime_path,
        "steam/i386/lib/i386-linux-gnu",
        "steam/i386/lib",
        "steam/i386/usr/lib/i386-linux-gnu",
        "steam/i386/usr/lib",
    ]

    if LINUX_SYSTEM.is_64_bit:
        lutris_runtime_path = "%s-x86_64" % version
        runtime_paths += [
            lutris_runtime_path,
            "steam/amd64/lib/x86_64-linux-gnu",
            "steam/amd64/lib",
            "steam/amd64/usr/lib/x86_64-linux-gnu",
            "steam/amd64/usr/lib",
        ]

    paths = []
    if prefer_system_libs:
        if wine_path:
            paths += get_winelib_paths(wine_path)
        paths += list(LINUX_SYSTEM.iter_lib_folders())
    # Then resolve absolute paths for the runtime
    paths += [os.path.join(settings.RUNTIME_DIR, path) for path in runtime_paths]
    return paths


def get_paths(
    version: Optional[str] = None, prefer_system_libs: bool = True, wine_path: Optional[str] = None
) -> List[str]:
    """Return a list of paths containing the runtime libraries."""
    if not RUNTIME_DISABLED:
        paths = get_runtime_paths(version=version, prefer_system_libs=prefer_system_libs, wine_path=wine_path)
    else:
        paths = []
    # Put existing LD_LIBRARY_PATH at the end
    if os.environ.get("LD_LIBRARY_PATH"):
        paths.append(os.environ["LD_LIBRARY_PATH"])
    return paths


class ComponentUpdater:
    (PENDING, DOWNLOADING, EXTRACTING, COMPLETED) = list(range(4))

    status_formats = {
        PENDING: _("Updating %s"),
        DOWNLOADING: _("Downloading %s"),
        EXTRACTING: _("Extracting %s"),
        COMPLETED: _("Updated %s"),
    }

    @property
    def name(self) -> str:
        raise NotImplementedError

    @property
    def should_update(self) -> bool:
        """True if this update should be installed; false to discard it."""
        return True

    def install_update(self, updater: "RuntimeUpdater") -> None:
        """Performs the update; runs on a worker thread, and returns when complete.
        However, some updates spawn a further thread to extract in parallel with the
        next update even after this."""
        raise NotImplementedError

    def join(self):
        """Blocks until the update is entirely complete; used when install_update() spawns
        a thread to finish up. We call this on each update before closing the download queue."""

    def get_progress(self) -> ProgressInfo:
        """Returns the current progress for the updater as it runs; called from the main thread."""
        return ProgressInfo.ended()


class RuntimeUpdater:
    """Class handling the runtime updates"""

    def __init__(self, force: bool = False):
        if RUNTIME_DISABLED:
            logger.warning("Runtime disabled by environment variable. Re-enable runtime before submitting issues.")
            self.update_runtime = False
        elif force:
            self.update_runtime = True
        else:
            self.update_runtime = settings.read_bool_setting("auto_update_runtime", default=True)

            if not self.update_runtime:
                logger.warning("Runtime updates are disabled. This configuration is not supported.")

            if not check_stale_runtime_versions():
                self.update_runtime = False

        if self.has_updates:
            self.runtime_versions = download_runtime_versions()
        else:
            self.runtime_versions = get_runtime_versions()

    @property
    def has_updates(self):
        return self.update_runtime

    def create_component_updaters(self) -> List[ComponentUpdater]:
        """Creates the component updaters that need to be applied and returns them in a list.

        This tests each to see if it should be used and returns only those you should install.
        It returns an empty list if has_updates is false.

        This method also downloads fresh runner versions on each call, so we call this on a
        worker thread, instead of blocking the UI."""
        if not self.runtime_versions:
            return []

        updaters: List[ComponentUpdater] = []

        if self.update_runtime:
            updaters += self._get_runtime_updaters(self.runtime_versions)

        return [u for u in updaters if u.should_update]

    def check_client_versions(self) -> Optional[str]:
        """Checks if the client is of an old version and no longer supported; this can be blocked
        with an env-var, and can be temporarily ignored as well, but if we're out of date, then
        this method returns the version of Lutris that is required. If we're good, or we are ignoring
        the problem, this returns None.

        I expect that most users will not discover the env-var, so they will be prompted
        on each new Lutris release."""
        if self.runtime_versions and not os.environ.get("LUTRIS_NO_CLIENT_VERSION_CHECK"):
            client_version = self.runtime_versions.get("client_version")
            if client_version:
                if parse_version(client_version) > parse_version(settings.VERSION):
                    ignored = settings.read_setting("ignored_supported_lutris_verison")
                    if not ignored or ignored != client_version:
                        return client_version

        return None

    @staticmethod
    def _get_runtime_updaters(runtime_versions: Dict[str, Any]) -> List[ComponentUpdater]:
        """Launch the update process"""
        updaters: List[ComponentUpdater] = []
        for name, remote_runtime in runtime_versions.get("runtimes", {}).items():
            if remote_runtime["architecture"] == "x86_64" and not LINUX_SYSTEM.is_64_bit:
                logger.debug("Skipping runtime %s for %s", name, remote_runtime["architecture"])
                continue

            try:
                url = remote_runtime.get("url")
                if url:
                    if "-proton" not in os.path.basename(url).casefold():
                        updaters.append(RuntimeExtractedComponentUpdater(remote_runtime))
                else:
                    updaters.append(RuntimeFilesComponentUpdater(remote_runtime))
            except Exception as ex:
                logger.exception("Unable to download %s: %s", name, ex)

        return updaters


class RuntimeComponentUpdater(ComponentUpdater):
    """A base class for component updates that use the timestamp from a runtime-info dict
    to decide when an update is required.

    These dicts are part of the 'versions.json' file, sent down from the server. They describe
    what versions of components are available from the server. This dict includes the modification
    date-time of the component, and we compare it to the mtime of the relevant directory.
    """

    def __init__(self, remote_runtime_info: Dict[str, Any]) -> None:
        self.remote_runtime_info = remote_runtime_info
        self.state = ComponentUpdater.PENDING
        # Versioned runtimes keep 1 version per folder
        self.versioned = bool(self.remote_runtime_info.get("versioned"))
        self.version = str(self.remote_runtime_info.get("version") or "") if self.versioned else ""

    @property
    def name(self) -> str:
        return self.remote_runtime_info["name"]

    @property
    def local_runtime_path(self) -> str:
        """Return the local path for the runtime directory where the component
        is kept."""
        return os.path.join(settings.RUNTIME_DIR, self.name)

    def install_update(self, updater: "RuntimeUpdater") -> None:
        raise NotImplementedError

    def get_progress(self) -> ProgressInfo:
        status_text = ComponentUpdater.status_formats[self.state] % self.name

        if self.state == ComponentUpdater.COMPLETED:
            return ProgressInfo.ended(status_text)

        if self.state == ComponentUpdater.PENDING:
            return ProgressInfo(0.0, status_text)

        return ProgressInfo(0, status_text)

    def get_updated_at(self) -> time.struct_time:
        """Return the modification date of the runtime folder"""
        return time.gmtime(os.path.getmtime(self.local_runtime_path))

    def set_updated_at(self) -> None:
        """Set the creation and modification time to now"""
        if not system.path_exists(self.local_runtime_path):
            logger.error("No local runtime path in %s", self.local_runtime_path)
            return
        os.utime(self.local_runtime_path)

    @property
    def should_update(self) -> bool:
        """Determine if the current runtime should be updated"""
        if self.versioned:
            return not system.path_exists(os.path.join(self.local_runtime_path, self.version))

        try:
            local_updated_at = self.get_updated_at()
        except FileNotFoundError:
            return True

        remote_updated_at = get_time_from_api_date(self.remote_runtime_info["created_at"])

        if local_updated_at and local_updated_at >= remote_updated_at:
            return False

        logger.debug(
            "Runtime %s locally updated on %s, remote created on %s)",
            self.name,
            time.strftime("%c", local_updated_at),
            time.strftime("%c", remote_updated_at),
        )
        return True


class RuntimeExtractedComponentUpdater(RuntimeComponentUpdater):
    """Component updater that downloads and extracts an archive."""

    def __init__(self, remote_runtime_info: Dict[str, Any]) -> None:
        super().__init__(remote_runtime_info)
        self.url = remote_runtime_info["url"]
        self.downloader: Downloader = None
        self.complete_event = threading.Event()

    def get_progress(self) -> ProgressInfo:
        progress_info = super().get_progress()

        if self.downloader and not progress_info.has_ended:
            return ProgressInfo(self.downloader.progress_fraction, progress_info.label_markup, self.downloader.cancel)

        return progress_info

    @property
    def archive_path(self):
        """This is the path where the archive is downloaded, before being extracted."""
        return os.path.join(settings.RUNTIME_DIR, os.path.basename(self.url))

    def install_update(self, updater: RuntimeUpdater) -> None:
        self.state = ComponentUpdater.DOWNLOADING
        self.complete_event.clear()

        archive_path = self.archive_path
        self.downloader = Downloader(self.url, archive_path, overwrite=True)
        self.downloader.start()
        self.downloader.join()
        self.downloader = None

        AsyncCall(self._install, self._install_cb, archive_path)

    def join(self):
        self.complete_event.wait()

    def _complete(self):
        self.state = ComponentUpdater.COMPLETED
        self.complete_event.set()

    def _install(self, path: str):
        """Finishes the installation after download, on a worker thread. This extracts
        the archive and downloads the versions file for it, the marks the update complete
        so join() above will be unblocked."""
        try:
            self._extract(path)

            self.set_updated_at()
            if self.name in DLL_MANAGERS:
                manager = DLL_MANAGERS[self.name]()
                manager.fetch_versions()
        finally:
            self._complete()

    def _install_cb(self, _completed: bool, error: Exception):
        if error:
            logger.error("Runtime update failed: %s", error)

    def _extract(self, path: str) -> None:
        """Actions taken once a runtime is downloaded

        Arguments:
            path: local path to the runtime archive, or None on download failure
        """
        if not path:
            return

        stats = os.stat(path)
        if not stats.st_size:
            logger.error("Download failed: file %s is empty, Deleting file.", path)
            os.unlink(path)
            return
        directory, _filename = os.path.split(path)

        # Determine the destination path
        dest_path = os.path.join(directory, self.name)

        if self.versioned:
            dest_path = os.path.join(dest_path, self.version)
        else:
            # Delete the existing runtime path
            system.delete_folder(dest_path)
        # Extract the runtime archive
        self.state = ComponentUpdater.EXTRACTING
        archive_path, _destination_path = extract_archive(path, dest_path, merge_single=True)
        os.unlink(archive_path)


class RuntimeFilesComponentUpdater(RuntimeComponentUpdater):
    """Component updaters that downloads a set of files described by the server
    individually."""

    def install_update(self, updater: RuntimeUpdater) -> None:
        """Download a runtime item by individual components. Used for icons only at the moment"""
        self.state = ComponentUpdater.DOWNLOADING
        components = self._get_runtime_components()
        downloads = []
        for component in components:
            modified_at = get_time_from_api_date(component["modified_at"])
            if not self._should_update_component(component["filename"], modified_at):
                continue
            downloads.append(component)

        with concurrent.futures.ThreadPoolExecutor(max_workers=8) as executor:
            future_downloads = {
                executor.submit(self._download_component, component): component["filename"] for component in downloads
            }
            for future in concurrent.futures.as_completed(future_downloads):
                if not future.cancelled() and future.exception():
                    expected_filename = future_downloads[future]
                    logger.warning("Failed to get '%s': %s", expected_filename, future.exception())
        self.state = ComponentUpdater.COMPLETED

    def _should_update_component(self, filename: str, remote_modified_at: time.struct_time) -> bool:
        """Should an individual component be updated?"""
        file_path = os.path.join(self.local_runtime_path, filename)
        if not system.path_exists(file_path):
            return True
        locally_modified_at = time.gmtime(os.path.getmtime(file_path))
        if locally_modified_at >= remote_modified_at:
            return False
        return True

    def _get_runtime_components(self) -> List[Dict[str, Any]]:
        """Fetch individual runtime files for a component"""
        request = http.Request(settings.RUNTIME_URL + "/" + self.name)
        try:
            response = request.get()
        except http.HTTPError as ex:
            logger.error("Failed to get components: %s", ex)
            return []
        if not response.json:
            return []
        return response.json.get("components", [])

    def _download_component(self, component: Dict[str, Any]) -> None:
        """Download an individual file from a runtime item"""
        file_path = os.path.join(self.local_runtime_path, component["filename"])
        http.download_file(component["url"], file_path)



import os
from gettext import gettext as _
from typing import Callable, Iterable, Optional

from gi.repository import Gio, GLib, GObject

from lutris.util.jobs import schedule_at_idle
from lutris.util.log import logger

PORTAL_BUS_NAME = "org.freedesktop.portal.Desktop"
PORTAL_OBJECT_PATH = "/org/freedesktop/portal/desktop"


class TrashPortal(GObject.Object):
    portal_interface = "org.freedesktop.portal.Trash"
    _dbus_proxy = None

    CompletionFunction = Callable[[], None]
    ErrorFunction = Callable[[Exception], None]

    def __init__(
        self,
        file_paths: Iterable[str],
        completion_function: Optional[CompletionFunction] = None,
        error_function: Optional[ErrorFunction] = None,
    ):
        super().__init__()
        self.file_paths = list(file_paths)
        self.completion_function = completion_function
        self.error_function = error_function
        Gio.DBusProxy.new_for_bus(
            Gio.BusType.SESSION,
            Gio.DBusProxyFlags.NONE,
            None,
            PORTAL_BUS_NAME,
            PORTAL_OBJECT_PATH,
            self.portal_interface,
            None,
            self._new_for_bus_cb,
        )

    def _new_for_bus_cb(self, obj, result):
        proxy = obj.new_for_bus_finish(result)
        if proxy:
            self._dbus_proxy = proxy
            self.trash_file()

    def trash_file(self):
        try:
            fds_in = Gio.UnixFDList.new()

            for file_path in self.file_paths:
                flags = os.O_RDONLY | os.O_PATH | os.O_CLOEXEC
                # You'd think you could use O_NOFOLLOW for any file, but
                # I find TrashFile fails. We don't want to trash the link target
                # in any case.
                if os.path.islink(file_path):
                    flags |= os.O_NOFOLLOW
                file_handle = os.open(file_path, flags)
                fds_in.append(file_handle)

            self._dbus_proxy.call_with_unix_fd_list(
                "TrashFile",
                GLib.Variant.new_tuple(
                    GLib.Variant.new_handle(0),
                ),
                Gio.DBusCallFlags.NONE,
                GObject.G_MAXINT,
                fds_in,
                None,
                self._call_cb,
            )
        except Exception as ex:
            self.report_error(ex)

    def _call_cb(self, obj, result):
        values = obj.call_finish(result)
        if values:
            result = values[0]
            if result == 0:
                if len(self.file_paths) == 1:
                    message = (
                        _("'%s' could not be moved to the trash. You will need to delete it yourself.")
                        % self.file_paths[0]
                    )
                else:
                    message = _(
                        "The items could not be moved to the trash. You will need to delete them yourself:\n%s"
                    ) % "\n".join(self.file_paths)
                self.report_error(RuntimeError(message))
                return
        self.report_completion()

    def report_error(self, error: Exception) -> None:
        if self.error_function:
            schedule_at_idle(self.error_function, error)
        else:
            logger.exception("Failed to trash %s: %s", ", ".join(self.file_paths), error)

    def report_completion(self):
        if self.completion_function:
            schedule_at_idle(self.completion_function)



"""Dialog used to install versions of a runner"""

# pylint: disable=no-member
import gettext
import os
import re
from collections import defaultdict
from gettext import gettext as _

from gi.repository import Gtk

from lutris import api, settings
from lutris.api import format_runner_version, parse_version_architecture
from lutris.database.games import get_games_by_runner
from lutris.game import Game
from lutris.gui.dialogs import ErrorDialog, ModelessDialog, display_error
from lutris.gui.widgets.utils import has_stock_icon
from lutris.util import jobs, system
from lutris.util.downloader import Downloader
from lutris.util.extract import extract_archive
from lutris.util.jobs import schedule_repeating_at_idle
from lutris.util.log import logger


def get_runner_path(runner_directory, version, arch):
    """Return the local path where the runner is/will be installed"""
    info = {"version": version, "architecture": arch}
    return os.path.join(runner_directory, format_runner_version(info))


def get_installed_versions(runner_directory):
    """List versions available locally"""
    if not os.path.exists(runner_directory):
        return set()
    return {parse_version_architecture(p) for p in os.listdir(runner_directory)}


def get_usage_stats(runner_name):
    """Return the usage for each version"""
    runner_games = get_games_by_runner(runner_name)
    version_usage = defaultdict(list)
    for db_game in runner_games:
        if not db_game["installed"]:
            continue
        game = Game(db_game["id"])
        version = game.config.runner_config["version"]
        version_usage[version].append(db_game["id"])
    return version_usage


class ShowAppsDialog(ModelessDialog):
    def __init__(self, title, parent, runner_name, runner_version):
        super().__init__(title, parent, border_width=10)
        self.runner_name = runner_name
        self.runner_version = runner_version
        self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.set_default_size(600, 400)
        self.apps = []
        label = Gtk.Label.new(_("Showing games using %s") % runner_version)
        self.vbox.add(label)
        scrolled_listbox = Gtk.ScrolledWindow()
        self.listbox = Gtk.ListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        scrolled_listbox.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_listbox.set_shadow_type(Gtk.ShadowType.ETCHED_OUT)
        scrolled_listbox.add(self.listbox)
        self.vbox.pack_start(scrolled_listbox, True, True, 14)
        self.show_all()
        jobs.AsyncCall(self.load_apps, self.on_apps_loaded)

    def load_apps(self):
        runner_games = get_games_by_runner(self.runner_name)
        for db_game in runner_games:
            if not db_game["installed"]:
                continue
            game = Game(db_game["id"])
            version = game.config.runner_config["version"]
            if version != self.runner_version:
                continue
            self.apps.append(game)

    def on_apps_loaded(self, _result, _error):
        for app in self.apps:
            row = Gtk.ListBoxRow(visible=True)
            hbox = Gtk.Box(visible=True, orientation=Gtk.Orientation.HORIZONTAL)
            lbl_game = Gtk.Label(app.name, visible=True)
            lbl_game.set_halign(Gtk.Align.START)
            hbox.pack_start(lbl_game, True, True, 5)
            row.add(hbox)
            self.listbox.add(row)


class RunnerInstallDialog(ModelessDialog):
    """Dialog displaying available runner version and downloads them"""

    COL_VER = 0
    COL_ARCH = 1
    COL_URL = 2
    COL_INSTALLED = 3
    COL_PROGRESS = 4
    COL_USAGE = 5

    def __init__(self, title, parent, runner):
        super().__init__(title, parent, 0, border_width=10)
        self.ok_button = self.add_default_button(Gtk.STOCK_OK, Gtk.ResponseType.OK)
        self.runner_name = runner.name
        self.runner_directory = runner.directory
        self.runner_info = {}
        self.runner_store = []
        self.installing = {}
        self.set_default_size(640, 480)
        self.runners = []
        self.listbox = None

        label = Gtk.Label.new(_("Waiting for response from %s") % settings.SITE_URL)
        self.vbox.pack_start(label, False, False, 18)

        spinner = Gtk.Spinner(visible=True)
        spinner.start()
        self.vbox.pack_start(spinner, False, False, 18)

        self.show_all()

        jobs.AsyncCall(self.fetch_runner_versions, self.runner_fetch_cb, self.runner_name, self.runner_directory)

    @staticmethod
    def fetch_runner_versions(runner_name, runner_directory):
        runner_info = api.get_runners(runner_name)
        runner_info["runner_name"] = runner_name
        runner_info["runner_directory"] = runner_directory
        remote_versions = {(v["version"], v["architecture"]) for v in runner_info["versions"]}
        local_versions = get_installed_versions(runner_directory)
        for local_version in local_versions - remote_versions:
            runner_info["versions"].append(
                {
                    "version": local_version[0],
                    "architecture": local_version[1],
                    "url": "",
                }
            )

        return runner_info, RunnerInstallDialog.fetch_runner_store(runner_info)

    @staticmethod
    def fetch_runner_store(runner_info):
        """Return a list populated with the runner versions"""
        runner_store = []
        runner_name = runner_info["runner_name"]
        runner_directory = runner_info["runner_directory"]
        version_usage = get_usage_stats(runner_name)
        ordered = sorted(runner_info["versions"], key=RunnerInstallDialog.get_version_sort_key)
        for version_info in reversed(ordered):
            is_installed = os.path.exists(
                get_runner_path(runner_directory, version_info["version"], version_info["architecture"])
            )
            games_using = version_usage.get("%(version)s-%(architecture)s" % version_info)
            runner_store.append(
                {
                    "version": version_info["version"],
                    "architecture": version_info["architecture"],
                    "url": version_info["url"],
                    "is_installed": is_installed,
                    "progress": 0,
                    "game_count": len(games_using) if games_using else 0,
                }
            )
        return runner_store

    def runner_fetch_cb(self, result, error):
        """Clear the box and display versions from runner_info"""
        if self.vbox is None:
            return  # if the dialog has closed before the runner fetch completed

        if error:
            logger.error(error)
            ErrorDialog(_("Unable to get runner versions: %s") % error, parent=self)
            return

        self.runner_info, self.runner_store = result

        if not self.runner_info:
            ErrorDialog(_("Unable to get runner versions from lutris.net"), parent=self)
            return

        for child_widget in self.vbox.get_children():
            if child_widget.get_name() not in "GtkBox":
                child_widget.destroy()

        label = Gtk.Label.new(_("%s version management") % self.runner_info["name"])
        self.vbox.add(label)
        self.installing = {}

        scrolled_listbox = Gtk.ScrolledWindow()
        self.listbox = Gtk.ListBox()
        self.listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        scrolled_listbox.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_listbox.set_shadow_type(Gtk.ShadowType.ETCHED_OUT)
        scrolled_listbox.add(self.listbox)
        self.vbox.pack_start(scrolled_listbox, True, True, 14)
        self.show_all()
        self.populate_listboxrows()

    def populate_listboxrows(self):
        icon_name = "software-installed-symbolic" if has_stock_icon("software-installed-symbolic") else "wine-symbolic"

        for runner in self.runner_store:
            row = Gtk.ListBoxRow()
            row.runner = runner
            hbox = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL)
            row.hbox = hbox

            icon = Gtk.Image.new_from_icon_name(icon_name, Gtk.IconSize.MENU)
            icon.set_visible(runner["is_installed"])
            icon_container = Gtk.Box()
            icon_container.set_size_request(16, 16)
            icon_container.pack_start(icon, False, False, 0)

            hbox.pack_start(icon_container, False, True, 0)
            row.icon = icon

            lbl_version = Gtk.Label(runner["version"])
            lbl_version.set_max_width_chars(20)
            lbl_version.set_property("width-chars", 20)
            lbl_version.set_halign(Gtk.Align.START)
            hbox.pack_start(lbl_version, False, False, 5)

            arch_label = Gtk.Label(runner["architecture"])
            arch_label.set_max_width_chars(8)
            arch_label.set_halign(Gtk.Align.START)
            hbox.pack_start(arch_label, False, True, 5)

            install_progress = Gtk.ProgressBar()
            install_progress.set_show_text(True)
            hbox.pack_end(install_progress, True, True, 5)
            row.install_progress = install_progress

            if runner["is_installed"]:
                # Check if there are apps installed, if so, show the view apps button
                app_count = runner["game_count"] or 0
                if app_count > 0:
                    usage_button_text = gettext.ngettext("View %d game", "View %d games", app_count) % app_count

                    usage_button = Gtk.LinkButton.new_with_label(usage_button_text)
                    usage_button.set_valign(Gtk.Align.CENTER)
                    usage_button.connect("clicked", self.on_show_apps_usage, row)
                    hbox.pack_end(usage_button, False, True, 2)

            button = Gtk.Button()
            button.set_size_request(100, -1)
            hbox.pack_end(button, False, True, 0)
            hbox.reorder_child(button, 0)
            row.install_uninstall_cancel_button = button
            row.handler_id = None

            row.add(hbox)
            self.listbox.add(row)
            row.show_all()
            self.update_listboxrow(row)

    def update_listboxrow(self, row):
        row.install_progress.set_visible(False)
        self.ok_button.set_sensitive(not self.installing)

        runner = row.runner
        icon = row.icon
        icon.set_visible(runner["is_installed"])

        button = row.install_uninstall_cancel_button
        style_context = button.get_style_context()
        if row.handler_id is not None:
            button.disconnect(row.handler_id)
            row.handler_id = None
        if runner["version"] in self.installing:
            style_context.remove_class("destructive-action")
            button.set_label(_("Cancel"))
            handler_id = button.connect("clicked", self.on_cancel_install, row)
        else:
            if runner["is_installed"]:
                style_context.add_class("destructive-action")
                button.set_label(_("Uninstall"))
                handler_id = button.connect("clicked", self.on_uninstall_runner, row)
            elif not runner["url"]:
                # Local-only version without download URL
                style_context.remove_class("destructive-action")
                button.set_label(_("Unavailable"))
                button.set_sensitive(False)
                handler_id = None
            else:
                style_context.remove_class("destructive-action")
                button.set_label(_("Install"))
                button.set_sensitive(True)
                handler_id = button.connect("clicked", self.on_install_runner, row)

        row.install_uninstall_cancel_button = button
        row.handler_id = handler_id

    def on_show_apps_usage(self, _button, row):
        """Return grid with games that uses this wine version"""
        runner = row.runner
        runner_version = "%s-%s" % (runner["version"], runner["architecture"])
        dialog = ShowAppsDialog(_("Wine version usage"), self.get_toplevel(), self.runner_name, runner_version)
        dialog.run()

    @staticmethod
    def get_version_sort_key(version):
        """Generate a sorting key that sorts first on the version number part of the version,
        and which breaks the version number into its components, which are parsed as integers"""
        raw_version = version["version"]
        # Extract version numbers from the end of the version string.
        # We look for things like xx-7.2 or xxx-4.3-2. A leading period
        # will be part of the version, but a leading hyphen will not.
        match = re.search(r"^(.*?)\-?(\d[.\-\d]*)$", raw_version)
        if match:
            version_parts = [int(p) for p in match.group(2).replace("-", ".").split(".") if p]
            return version_parts, raw_version, version["architecture"]

        # If we fail to extract the version, we'll wind up sorting this one to the end.
        return [], raw_version, version["architecture"]

    def get_dest_path(self, runner):
        """Return temporary path where the runners should be downloaded to"""
        return os.path.join(settings.CACHE_DIR, os.path.basename(runner["url"]))

    def on_cancel_install(self, widget, row):
        self.cancel_install(row)

    def cancel_install(self, row):
        """Cancel the installation of a runner version"""
        runner = row.runner
        self.installing[runner["version"]].cancel()
        self.uninstall_runner(row)
        runner["progress"] = 0
        self.installing.pop(runner["version"])
        self.update_listboxrow(row)

    def on_uninstall_runner(self, widget, row):
        self.uninstall_runner(row)

    def uninstall_runner(self, row):
        """Uninstall a runner version"""
        runner = row.runner
        version = runner["version"]
        arch = runner["architecture"]
        runner_path = get_runner_path(self.runner_directory, version, arch)

        def on_complete():
            runner["is_installed"] = False
            if self.runner_name == "wine":
                logger.debug("Clearing wine version cache")
                from lutris.util.wine.wine import clear_wine_version_cache

                clear_wine_version_cache()
            self.update_listboxrow(row)

        def on_error(error):
            logger.exception("Runner '%s (%s)' uninstall failed: %s", self.runner_name, version, error)
            display_error(error, parent=self)

        system.remove_folder(runner_path, completion_function=on_complete, error_function=on_error)

    def on_install_runner(self, _widget, row):
        self.install_runner(row)

    def install_runner(self, row):
        """Download and install a runner version"""
        runner = row.runner
        row.install_progress.set_fraction(0.0)
        dest_path = self.get_dest_path(runner)
        url = runner["url"]
        version = runner["version"]
        if not url:
            ErrorDialog(_("Version %s is no longer available") % version, parent=self)
            return
        downloader = Downloader(url, dest_path, overwrite=True)
        schedule_repeating_at_idle(self.get_progress, downloader, row, interval_seconds=0.1)
        self.installing[version] = downloader
        downloader.start()
        self.update_listboxrow(row)

    def get_progress(self, downloader, row) -> bool:
        """Update progress bar with download progress"""
        runner = row.runner
        if downloader.state == downloader.CANCELLED:
            return False
        if downloader.state == downloader.ERROR:
            self.cancel_install(row)
            return False
        row.install_progress.show()
        downloader.check_progress()
        percent_downloaded = downloader.progress_percentage
        if percent_downloaded >= 1:
            runner["progress"] = percent_downloaded
            row.install_progress.set_fraction(percent_downloaded / 100)
        else:
            runner["progress"] = 1
            row.install_progress.pulse()
            row.install_progress.set_text = _("Downloading…")
        if downloader.state == downloader.COMPLETED:
            runner["progress"] = 99
            row.install_progress.set_text = _("Extracting…")
            self.on_runner_downloaded(row)
            return False
        return True

    def progress_pulse(self, row) -> bool:
        runner = row.runner
        row.install_progress.pulse()
        return not runner["is_installed"]

    def on_runner_downloaded(self, row):
        """Handler called when a runner version is downloaded"""
        runner = row.runner
        version = runner["version"]
        architecture = runner["architecture"]
        logger.debug("Runner %s for %s has finished downloading", version, architecture)
        src = self.get_dest_path(runner)
        dst = get_runner_path(self.runner_directory, version, architecture)
        schedule_repeating_at_idle(self.progress_pulse, row, interval_seconds=0.1)
        jobs.AsyncCall(self.extract, self.on_extracted, src, dst, row)

    @staticmethod
    def extract(src, dst, row):
        """Extract a runner archive to a destination"""
        extract_archive(src, dst)
        return src, row

    def on_extracted(self, row_info, error):
        """Called when a runner archive is extracted"""
        if error or not row_info:
            ErrorDialog(_("Failed to retrieve the runner archive"), parent=self)
            return
        src, row = row_info
        runner = row.runner
        os.remove(src)
        runner["progress"] = 0
        runner["is_installed"] = True
        self.installing.pop(runner["version"])
        row.install_progress.set_text = ""
        row.install_progress.set_fraction(0.0)
        row.install_progress.hide()
        self.update_listboxrow(row)
        if self.runner_name == "wine":
            logger.debug("Clearing wine version cache")
            from lutris.util.wine.wine import clear_wine_version_cache

            clear_wine_version_cache()

    def on_response(self, dialog, response: Gtk.ResponseType) -> None:
        if self.installing:
            return
        super().on_response(dialog, response)



# pylint: disable=no-member
"""Hardware driver related utilities

Everything in this module should rely on /proc or /sys only, no executable calls
"""

import os
import re
from typing import Dict, Iterable, List

from lutris.util import cache_single
from lutris.util.graphics.glxinfo import GlxInfo
from lutris.util.log import logger
from lutris.util.system import read_process_output

MIN_RECOMMENDED_NVIDIA_DRIVER = 515


@cache_single
def get_nvidia_driver_info() -> Dict[str, str]:
    """Return information about Nvidia drivers"""
    version_file_path = "/proc/driver/nvidia/version"

    def read_from_proc() -> Dict[str, str]:
        try:
            if not os.path.exists(version_file_path):
                return {}
            with open(version_file_path, encoding="utf-8") as version_file:
                content = version_file.readlines()
        except PermissionError:
            # MAC systems (selinux, apparmor) may block access to files in /proc.
            # If this happens, we may still be able to retrieve the info by
            # other means, but need additional validation.
            logger.info("Could not access %s. Falling back to glxinfo.", version_file_path)
            return {}
        except OSError as e:
            logger.warning(
                "Unexpected error when accessing %s. Falling back to glxinfo.",
                version_file_path,
                exc_info=e,
            )
            return {}

        try:
            nvrm_version = content[0].split(": ")[1].strip().split()
            if "Open" in nvrm_version:
                return {
                    "vendor": nvrm_version[0],
                    "platform": nvrm_version[1],
                    "arch": nvrm_version[6],
                    "version": nvrm_version[7],
                }
            return {
                "vendor": nvrm_version[0],
                "platform": nvrm_version[1],
                "arch": nvrm_version[2],
                "version": nvrm_version[5],
                "date": " ".join(nvrm_version[6:]),
            }
        except IndexError as ex:
            logger.warning("Unable to parse %s. Falling back to glxinfo: %s", version_file_path, ex)
            return {}

    def invoke_glxinfo() -> Dict[str, str]:
        glx_info = GlxInfo()
        platform = read_process_output(["uname", "-s"])
        arch = read_process_output(["uname", "-m"])
        vendor = glx_info.opengl_vendor  # type: ignore[attr-defined]
        if "nvidia" not in vendor.lower():
            raise RuntimeError("Expected NVIDIA vendor information, received %s." % vendor)
        return {
            "vendor": vendor,
            "platform": platform,
            "arch": arch,
            "version": glx_info.opengl_version.rsplit(maxsplit=1)[-1],  # type: ignore[attr-defined]
        }

    try:
        from_proc = read_from_proc()
        if from_proc:
            return from_proc
    except Exception as ex:
        logger.exception("Unable to read from '%s': %s", version_file_path, ex)

    return invoke_glxinfo()


def get_nvidia_gpu_ids() -> List[str]:
    """Return the list of Nvidia GPUs"""
    gpus_dir = "/proc/driver/nvidia/gpus"
    try:
        return os.listdir(gpus_dir)
    except PermissionError:
        logger.info("Permission denied to %s. Using lspci instead.", gpus_dir)
    except OSError as e:
        logger.warning("Unexpected error accessing %s. Using lspci instead.", gpus_dir, exc_info=e)
    # 10de is NVIDIA's vendor ID, 0300 gets you video controllers.
    values = read_process_output(["lspci", "-D", "-n", "-d", "10de::0300"]).splitlines()
    return [line.split(maxsplit=1)[0] for line in values]


def get_nvidia_gpu_info(gpu_id: str) -> Dict[str, str]:
    """Return details about a GPU"""
    gpu_info_file = f"/proc/driver/nvidia/gpus/{gpu_id}/information"
    try:
        with open(gpu_info_file, encoding="utf-8") as info_file:
            content = info_file.readlines()
    except PermissionError:
        logger.info("Permission denied to %s. Detecting with lspci.", gpu_info_file)
    except OSError as e:
        logger.warning(
            "Unexpected error accessing %s. Detecting with lspci",
            gpu_info_file,
            exc_info=e,
        )
    else:
        info = {}
        for line in content:
            key, value = line.split(":", 1)
            info[key] = value.strip()
        return info
    return get_lspci_nvidia_gpu_info(gpu_id)


def get_lspci_nvidia_gpu_info(gpu_id: str) -> Dict[str, str]:
    lspci_data = read_process_output(["lspci", "-v", "-s", gpu_id])
    model_info = re.search(r"NVIDIA Corporation \w+ \[(.+?)\]", lspci_data)
    if model_info:
        model = model_info.group(1)
    else:
        logger.error("Could not detect NVIDIA GPU model.")
        model = "Unknown"
    irq_info = re.search("IRQ ([0-9]+)", lspci_data)
    if irq_info:
        irq = irq_info.group(1)
    else:
        logger.error("Could not detect GPU IRQ information.")
        irq = None

    info = {
        "Model": f"NVIDIA {model}",
        "IRQ": str(irq),
        "Bus Location": str(gpu_id),
    }
    for line in lspci_data.splitlines():
        if ":" not in line:
            continue
        key, value = line.split(":", 1)
        info[key.strip()] = value.strip()
    return info


def is_nvidia() -> bool:
    """Return true if the Nvidia drivers are currently in use.

    Note: This function may not detect use of the nouveau drivers.
    """

    try:
        return os.path.exists("/proc/driver/nvidia")
    except OSError:
        logger.info("Could not determine whether /proc/driver/nvidia exists. Falling back to alternative method")
    try:
        with open("/proc/modules", encoding="utf-8") as f:
            modules = f.read()
        return bool(re.search(r"^nvidia ", modules, flags=re.MULTILINE))
    except OSError:
        logger.error("Could not access /proc/modules to find the Nvidia drivers. Nvidia card may not be detected.")
    glx_info = GlxInfo()
    return "NVIDIA" in glx_info.opengl_vendor  # type: ignore[attr-defined]


def get_gpu_cards() -> Iterable[str]:
    """Return GPUs connected to the system"""
    if not os.path.exists("/sys/class/drm"):
        logger.error("No GPU available on this system!")
        return

    try:
        cardlist = os.listdir("/sys/class/drm/")
    except PermissionError:
        logger.error("Your system does not allow reading from /sys/class/drm, no GPU detected.")
        return

    for cardname in cardlist:
        if re.match(r"^card\d$", cardname):
            yield cardname


def get_gpu_info(card: str) -> Dict[str, str]:
    """Return information about a GPU"""
    infos = {"DRIVER": "", "PCI_ID": "", "PCI_SUBSYS_ID": ""}
    try:
        with open(f"/sys/class/drm/{card}/device/uevent", encoding="utf-8") as card_uevent:
            content = card_uevent.readlines()
    except FileNotFoundError:
        logger.error("Unable to read driver information for card %s", card)
        return infos
    for line in content:
        key, value = line.split("=", 1)
        infos[key] = value.strip()
    return infos


def is_amd() -> bool:
    """Return true if the system uses the AMD driver"""
    for card in get_gpu_cards():
        if get_gpu_info(card)["DRIVER"] == "amdgpu":
            return True
    return False


def is_outdated() -> bool:
    if not is_nvidia():
        return False
    driver_info = get_nvidia_driver_info()
    driver_version = driver_info["version"]
    if not driver_version:
        logger.error("Failed to get Nvidia version")
        return True
    try:
        major_version = int(driver_version.split(".")[0])
    except (IndexError, ValueError) as ex:
        logger.exception("Failed to parse Nvidia version: %s", ex)
        return True
    return major_version < MIN_RECOMMENDED_NVIDIA_DRIVER



"""Controls execution of programs in separate shells"""

import os
from textwrap import dedent

from lutris import settings


def get_terminal_script(command, cwd, env):
    """Write command in a script file and run it.

    Running it from a file is likely the only way to set env vars only
    for the command (not for the terminal app).
    It's also the only reliable way to keep the term open when the
    game is quit.
    """
    script_path = os.path.join(settings.CACHE_DIR, "run_in_term.sh")
    env["TERM"] = "xterm"
    exported_environment = "\n".join('export %s="%s" ' % (key, value) for key, value in env.items())
    command = " ".join(['"%s"' % token for token in command])
    with open(script_path, "w", encoding="utf-8") as script_file:
        script_file.write(
            dedent(
                """\
            #!/bin/sh
            cd "%s"
            %s
            exec %s
            exit $?
            """
            )
            % (cwd, exported_environment, command)
        )
        os.chmod(script_path, 0o744)
    return script_path


def get_bash_rc_file(cwd, env, aliases=None):
    """Return a bash prompt configured with pre-defined environment variables and aliases"""
    script_path = os.path.join(settings.CACHE_DIR, "bashrc.sh")
    env["TERM"] = "xterm"
    exported_environment = "\n".join('export %s="%s"' % (key, value) for key, value in env.items())
    aliases = aliases or {}
    alias_commands = "\n".join('alias %s="%s"' % (key, value) for key, value in aliases.items())
    current_bashrc = os.path.expanduser("~/.bashrc")
    with open(script_path, "w", encoding="utf-8") as script_file:
        script_file.write(
            dedent(
                """
            . %s

            %s
            %s
            cd "%s"
            """
                % (
                    current_bashrc,
                    exported_environment,
                    alias_commands,
                    cwd,
                )
            )
        )
    return script_path


def get_shell_command(cwd, env, aliases=None):
    """Generates a scripts whichs opens a bash shell configured with given
    environment variables and aliases.
    """
    bashrc_file = get_bash_rc_file(cwd, env, aliases)
    return get_terminal_script(["bash", "--rcfile", bashrc_file], cwd, env)
